"""Module to contain field mappings for UM STASH codes to LFRic long names.

ORIGINAL CODE AT source=svn://fcm1/cma.xm_svn/contrib/trunk/irislib.
Please update the original code source as well as this file if changes required.
"""

# Set up UM STASH code to cube name mapping
# Form of entry is <stashcode>: (<long_name>, <grid>)
STASH_TO_LFRIC = {
    "m01s00i002": ("eastward_wind_at_cell_centres", "face"),
    "m01s00i003": ("northward_wind_at_cell_centres", "face"),
    "m01s00i004": ("air_potential_temperature", "face"),
    "m01s00i010": ("specific_humidity", "face"),
    "m01s00i012": ("specific_cloud_ice", "face"),
    "m01s00i023": ("grid_surface_snow_amount", "face"),
    "m01s00i024": ("grid_surface_temperature", "face"),
    "m01s00i025": ("boundary_layer_depth", "face"),
    "m01s00i026": ("surface_roughness_length", "face"),
    "m01s00i030": ("land_binary_mask", "face"),
    "m01s00i031": ("sea_ice_area_fraction", "face"),
    "m01s00i033": ("surface_altitude", "face"),
    "m01s00i150": ("upward_air_velocity_at_cell_interfaces", "face"),
    "m01s00i254": ("specific_cloud_liquid_water", "face"),
    "m01s00i265": ("area_cloud_fraction", "face"),
    "m01s00i266": ("bulk_cloud_fraction", "face"),
    "m01s00i267": ("liquid_cloud_fraction", "face"),
    "m01s00i268": ("frozen_cloud_fraction", "face"),
    "m01s00i389": ("dry_air_density", "face"),
    "m01s00i391": ("vapour_mixing_ratio", "face"),
    "m01s00i392": ("cloud_liquid_mixing_ratio", "face"),
    "m01s00i393": ("cloud_ice_mixing_ratio", "face"),
    "m01s00i394": ("rain_mixing_ratio", "face"),
    "m01s00i395": ("graupel_mixing_ratio", "face"),
    "m01s00i396": ("snow_mixing_ratio", "face"),
    "m01s00i407": ("pressure_at_cell_centres", "face"),
    "m01s00i408": ("pressure_at_cell_interfaces", "face"),
    "m01s00i409": ("surface_air_pressure", "face"),
    "m01s01i140": ("cosine_of_the_solar_zenith_angle", "face"),
    "m01s01i141": ("sunlit_fraction_of_the_timestep", "face"),
    "m01s01i142": ("cosine_of_the_solar_zenith_angle_radiative_timestep", "face"),
    "m01s01i143": ("sunlit_fraction_of_the_timestep_radiative_timestep", "face"),
    "m01s01i161": ("temperature_increment_from_sw_radiation", "face"),
    "m01s01i201": ("surface_net_shortwave_flux_radiative_timestep", "face"),
    "m01s01i202": ("surface_net_shortwave_flux", "face"),
    "m01s01i205": ("toa_upward_shortwave_flux", "face"),
    "m01s01i207": ("toa_direct_shortwave_flux", "face"),
    "m01s01i208": ("toa_upward_shortwave_flux_radiative_timestep", "face"),
    "m01s01i209": ("toa_upward_clear_shortwave_flux_radiative_timestep", "face"),
    "m01s01i210": ("surface_downward_clear_shortwave_flux_radiative_timestep", "face"),
    "m01s01i211": ("surface_upward_clear_shortwave_flux_radiative_timestep", "face"),
    "m01s01i215": ("surface_direct_shortwave_flux_radiative_timestep", "face"),
    "m01s01i216": ("surface_diffuse_shortwave_flux_radiative_timestep", "face"),
    "m01s01i217": ("upward_shortwave_flux_radiative_timestep", "face"),
    "m01s01i218": ("downward_shortwave_flux_radiative_timestep", "face"),
    "m01s01i219": ("upward_clear_shortwave_flux_radiative_timestep", "face"),
    "m01s01i220": ("downward_clear_shortwave_flux_radiative_timestep", "face"),
    "m01s01i235": ("surface_downward_shortwave_flux_radiative_timestep", "face"),
    "m01s01i254": (
        "weighted_warm_cloud_top_effective_radius_radiative_timestep",
        "face",
    ),
    "m01s01i255": ("warm_cloud_top_weight_radiative_timestep", "face"),
    "m01s02i161": ("temperature_increment_from_lw_radiation", "face"),
    "m01s02i201": ("surface_net_longwave_flux_radiative_timestep", "face"),
    "m01s02i204": ("total_column_cloud_fraction_radiative_timestep", "face"),
    "m01s02i205": ("toa_upward_longwave_flux_radiative_timestep", "face"),
    "m01s02i206": ("toa_upward_clear_longwave_flux_radiative_timestep", "face"),
    "m01s02i207": ("surface_downward_longwave_flux_radiative_timestep", "face"),
    "m01s02i208": ("surface_downward_clear_longwave_flux_radiative_timestep", "face"),
    "m01s02i217": ("upward_longwave_flux_radiative_timestep", "face"),
    "m01s02i218": ("downward_longwave_flux_radiative_timestep", "face"),
    "m01s02i219": ("upward_clear_longwave_flux_radiative_timestep", "face"),
    "m01s02i220": ("downward_clear_longwave_flux_radiative_timestep", "face"),
    "m01s02i298": ("aerosol_optical_depth_in_visible_radiative_timestep", "face"),
    "m01s02i308": ("liquid_cloud_mmr_radiative_timestep", "face"),
    "m01s02i309": ("ice_cloud_mmr_radiative_timestep", "face"),
    "m01s02i312": ("liquid_cloud_fraction_radiative_timestep", "face"),
    "m01s02i313": ("ice_cloud_fraction_radiative_timestep", "face"),
    "m01s02i321": ("calipso_low_cloud_mask", "face"),
    "m01s02i322": ("calipso_mid_cloud_mask", "face"),
    "m01s02i323": ("calipso_high_cloud_mask", "face"),
    "m01s02i325": ("calipso_cf_40_lvls_mask", "face"),
    "m01s02i330": ("sunlit_mask", "face"),
    "m01s02i337": ("isccp_ctp_tau", "face"),
    "m01s02i341": ("calipso_total_backscatter", "face"),
    "m01s02i344": ("calipso_low_cloud", "face"),
    "m01s02i345": ("calipso_mid_cloud", "face"),
    "m01s02i346": ("calipso_high_cloud", "face"),
    "m01s02i370": ("calipso_cfad_sr_40_lvls", "face"),
    "m01s02i473": ("calipso_cf_40_lvls_liq", "face"),
    "m01s02i474": ("calipso_cf_40_lvls_ice", "face"),
    "m01s02i475": ("calipso_cf_40_lvls_undet", "face"),
    "m01s03i025": ("boundary_layer_depth", "face"),
    "m01s03i181": ("temperature_increment_from_bl_scheme", "face"),
    "m01s03i182": ("vapour_increment_from_bl_scheme", "face"),
    "m01s03i183": ("liquid_water_increment_from_bl_scheme", "face"),
    "m01s03i184": ("frozen_water_increment_from_bl_scheme", "face"),
    "m01s03i185": ("eastward_wind_increment_from_bl_scheme", "face"),
    "m01s03i186": ("northward_wind_increment_from_bl_scheme", "face"),
    "m01s03i187": ("upward_air_velocity_increment_from_bl_scheme", "face"),
    "m01s03i192": ("bulk_cloud_fraction_increment_from_bl_scheme", "face"),
    "m01s03i193": ("liquid_cloud_fraction_increment_from_bl_scheme", "face"),
    "m01s03i194": ("frozen_cloud_fraction_increment_from_bl_scheme", "face"),
    "m01s03i208": ("lowest_layer_bulk_richardson_number", "face"),
    "m01s03i217": ("grid_surface_upward_sensible_heat_flux", "face"),
    "m01s03i219": ("surface_eastward_wind_stress", "face"),
    "m01s03i220": ("surface_northward_wind_stress", "face"),
    "m01s03i223": ("grid_surface_moisture_flux", "face"),
    "m01s03i225": ("eastward_wind_at_10m", "face"),
    "m01s03i226": ("northward_wind_at_10m", "face"),
    "m01s03i227": ("wind_speed_at_10m", "face"),
    "m01s03i234": ("grid_surface_upward_latent_heat_flux", "face"),
    "m01s03i236": ("temperature_at_screen_level", "face"),
    "m01s03i237": ("specific_humidity_at_screen_level", "face"),
    "m01s03i245": ("relative_humidity_at_screen_level", "face"),
    "m01s03i247": ("visibility_excluding precipitation_at_screen_level", "face"),
    "m01s03i248": ("fog_fraction_at_screen_level", "face"),
    "m01s03i258": ("surface_snow_melt_heat_flux", "face"),
    "m01s03i261": ("gross_primary_productivity", "face"),
    "m01s03i281": ("visibility_including_precipitation_at_screen_level", "face"),
    "m01s03i296": ("water_evaporation_flux_from_soil", "face"),
    "m01s03i297": ("grid_water_evaporation_flux_from_canopy", "face"),
    "m01s03i298": ("grid_surface_snow_sublimation_rate", "face"),
    "m01s03i304": ("turbulent_mixing_height", "face"),
    "m01s03i305": ("stable_boundary_layer_indicator", "face"),
    "m01s03i306": ("stratocumulus_over_stable_boundary_layer_indicator", "face"),
    "m01s03i307": ("wellmixed_boundary_layer_indicator", "face"),
    "m01s03i308": (
        "decoupled_stratocumulus_not_over_cumulus_boundary_layer_indicator",
        "face",
    ),
    "m01s03i309": (
        "decoupled_stratocumulus_over_cumulus_boundary_layer_indicator",
        "face",
    ),
    "m01s03i310": ("cumulus_capped_boundary_layer_indicator", "face"),
    "m01s03i340": ("shear_driven_boundary_layer_indicator", "face"),
    "m01s03i365": ("neutral_eastward_wind_at_10m", "face"),
    "m01s03i366": ("neutral_northward_wind_at_10m", "face"),
    "m01s03i367": ("neutral_wind_speed_at_10m", "face"),
    "m01s03i395": ("land_area_fraction", "face"),
    "m01s03i662": ("net_primary_productivity", "face"),
    "m01s04i142": (
        "tendency_of_atmosphere_water_vapor_content_due_to_pc2_checks",
        "face",
    ),
    "m01s04i143": (
        "tendency_of_atmosphere_cloud_liquid_water_content_due_to_pc2_checks",
        "face",
    ),
    "m01s04i144": (
        "tendency_of_atmosphere_cloud_ice_water_content_due_to_pc2_checks",
        "face",
    ),
    "m01s04i152": (
        "tendency_of_cloud_amount_in_atmosphere_layer_due_to_pc2_checks",
        "face",
    ),
    "m01s04i153": (
        "tendency_of_liquid_cloud_amount_in_atmosphere_layer_due_to_pc2_checks",
        "face",
    ),
    "m01s04i154": (
        "tendency_of_frozen_cloud_amount_in_atmosphere_layer_due_to_pc2_checks",
        "face",
    ),
    "m01s04i181": ("temperature_increment_due_to_microphysics", "face"),
    "m01s04i182": ("water_vapour_mixing_ratio_increment_due_to_microphysics", "face"),
    "m01s04i183": ("cloud_liquid__mixing_ratio_increment_due_to_microphysics", "face"),
    "m01s04i184": ("cloud_ice_mixing_ratio_increment_due_to_microphysics", "face"),
    "m01s04i189": ("rain_mixing_ratio_increment_due_to_microphysics", "face"),
    "m01s04i190": ("snow_mixing_ratio_increment_due_to_microphysics", "face"),
    "m01s04i191": ("graupel_mixing_ratio_increment_due_to_microphysics", "face"),
    "m01s04i192": ("bulk_cloud_volume_increment_due_to_microphysics", "face"),
    "m01s04i193": ("liquid_cloud_volume_increment_due_to_microphysics", "face"),
    "m01s04i194": ("frozen_cloud_volume_increment_due_to_microphysics", "face"),
    "m01s04i201": ("surface_microphysical_rainfall_amount", "face"),
    "m01s04i202": ("surface_microphysical_snowfall_amount", "face"),
    "m01s04i203": ("surface_microphysical_rainfall_rate", "face"),
    "m01s04i204": ("surface_microphysical_snowfall_rate", "face"),
    "m01s05i161": ("temperature_increment_from_convection", "face"),
    "m01s05i162": ("water_vapour_increment_from_convection", "face"),
    "m01s05i163": ("cloud_liquid_water_increment_from_convection", "face"),
    "m01s05i164": ("cloud_ice_water_increment_from_convection", "face"),
    "m01s05i172": ("bulk_cloud_volume_increment_from_convection", "face"),
    "m01s05i173": ("liquid_cloud_volume_increment_from_convection", "face"),
    "m01s05i174": ("frozen_cloud_volume_increment_from_convection", "face"),
    "m01s05i182": (
        "tendency_of_atmosphere_water_vapor_content_due_to_pc2_conv_coupling",
        "face",
    ),
    "m01s05i183": (
        "tendency_of_atmosphere_cloud_liquid_water_content_due_to_pc2_conv_coupling",
        "face",
    ),
    "m01s05i184": (
        "tendency_of_atmosphere_cloud_ice_water_content_due_to_pc2_conv_coupling",
        "face",
    ),
    "m01s05i185": ("u_increment_from_convection", "face"),
    "m01s05i186": ("v_increment_from_convection", "face"),
    "m01s05i187": (
        "potential_temperature_increment_from_convection_excluding_shallow_convection",
        "face",
    ),
    "m01s05i188": (
        "water_vapour_increment_from_convection_excluding_shallow_convection",
        "face",
    ),
    "m01s05i192": (
        "tendency_of_cloud_amount_in_atmosphere_layer_due_to_pc2_conv_coupling",
        "face",
    ),
    "m01s05i193": (
        "tendency_of_liquid_cloud_amount_in_atmosphere_layer_due_to_pc2_conv_coupling",
        "face",
    ),
    "m01s05i194": (
        "tendency_of_frozen_cloud_amount_in_atmosphere_layer_due_to_pc2_conv_coupling",
        "face",
    ),
    "m01s05i201": ("surface_convective_rainfall_amount", "face"),
    "m01s05i202": ("surface_convective_snowfall_amount", "face"),
    "m01s05i205": ("surface_convective_rainfall_rate", "face"),
    "m01s05i206": ("surface_convective_snowfall_rate", "face"),
    "m01s05i207": ("pressure_at_convective_cloud_base", "face"),
    "m01s05i208": ("pressure_at_convective_cloud_top", "face"),
    "m01s05i212": ("convective_cloud_amount", "face"),
    "m01s05i214": ("surface_rainfall_rate", "face"),
    "m01s05i215": ("surface_snowfall_rate", "face"),
    "m01s05i216": ("precipitation_rate", "face"),
    "m01s05i226": ("surface_precipitation_amount", "face"),
    "m01s05i246": ("convection_upward_massflux_half_levs", "face"),
    "m01s05i250": ("convection_upward_massflux", "face"),
    "m01s05i251": ("convection_downward_massflux", "face"),
    "m01s05i267": ("deep_convection_cfl_limited", "face"),
    "m01s05i268": ("midlevel_convection_cfl_limited", "face"),
    "m01s05i269": ("frequency_of_deep_convection", "face"),
    "m01s05i270": ("frequency_of_shallow_convection", "face"),
    "m01s05i272": ("frequency_of_midlevel_convection", "face"),
    "m01s05i277": ("deep_convective_precipitation_rate", "face"),
    "m01s05i278": ("shallow_convective_precipitation_rate", "face"),
    "m01s05i279": ("midlevel_convection_precipitation_rate", "face"),
    "m01s06i101": ("spectral_gwd_eastward_wind_stress", "face"),
    "m01s06i102": ("spectral_gwd_southward_wind_stress", "face"),
    "m01s06i103": ("spectral_gwd_westward_wind_stress", "face"),
    "m01s06i104": ("spectral_gwd_northward_wind_stress", "face"),
    "m01s06i105": ("spectral_gravity_wave_drag_U_wind_tendency", "face"),
    "m01s06i106": ("spectral_gravity_wave_drag_V_wind_tendency", "face"),
    "m01s06i201": ("orographic_drag_U_wind_stress", "face"),
    "m01s06i202": ("orographic_drag_V_wind_stress", "face"),
    "m01s06i207": ("orographic_gravity_wave_drag_U_wind_tendency", "face"),
    "m01s06i208": ("orographic_gravity_wave_drag_V_wind_tendency", "face"),
    "m01s06i223": ("orographic_gravity_wave_drag_U_wind_stress", "face"),
    "m01s06i224": ("orographic_gravity_wave_drag_V_wind_stress", "face"),
    "m01s08i208": ("mass_content_of_water_in_soil", "face"),
    "m01s08i209": ("grid_canopy_water_amount", "face"),
    "m01s08i223": ("mass_content_of_water_in_soil_layer", "face"),
    "m01s08i225": ("soil_temperature", "face"),
    "m01s08i229": (
        "mass_fraction_of_unfrozen_water_in_saturated_soil_moisture",
        "face",
    ),
    "m01s08i230": ("mass_fraction_of_frozen_water_in_saturated_soil_moisture", "face"),
    "m01s08i231": ("grid_surface_snow_melt_flux", "face"),
    "m01s08i233": ("grid_canopy_throughfall_flux", "face"),
    "m01s08i234": ("surface_runoff_flux", "face"),
    "m01s08i235": ("subsurface_runoff_flux", "face"),
    "m01s09i202": ("maximum_combined_cloud_amount_below_111m_asl", "face"),
    "m01s09i203": ("maximum_combined_cloud_amount_between_111_and_1949m_asl", "face"),
    "m01s09i204": ("maximum_combined_cloud_amount_between_1949_and_5574m_asl", "face"),
    "m01s09i205": ("maximum_combined_cloud_amount_between_5574_and_13608m_asl", "face"),
    "m01s09i210": (
        "cloud_base_altitude_asl_combined_cloud_amount_greater_than_2p5_okta",
        "face",
    ),
    "m01s09i216": ("combined_cloud_amount_random_overlap", "face"),
    "m01s09i217": ("combined_cloud_amount_maximum_random_overlap", "face"),
    "m01s09i233": (
        "ceilometer_filtered_combined_cloud_amount_maximum_random_overlap",
        "face",
    ),
    "m01s10i185": ("eastward_wind_increment_from_solver", "face"),
    "m01s10i186": ("northward_wind_increment_from_solver", "face"),
    "m01s10i187": ("vertical_air_velocity_increment_from_solver", "face"),
    "m01s12i185": ("eastward_wind_increment_from_advection", "face"),
    "m01s12i186": ("northward_wind_increment_from_advection", "face"),
    "m01s12i187": ("vertical_air_velocity_increment_from_advection", "face"),
    "m01s12i192": ("bulk_cloud_fraction_increment_from_advection", "face"),
    "m01s12i193": ("liquid_cloud_fraction_increment_from_advection", "face"),
    "m01s12i194": ("frozen_cloud_fraction_increment_from_advection", "face"),
    "m01s12i195": ("vapour_increment_from_advection", "face"),
    "m01s12i196": ("liquid_water_increment_from_advection", "face"),
    "m01s12i197": ("frozen_water_increment_from_advection", "face"),
    "m01s15i201": ("zonal_wind_at_pressure_levels", "face"),
    "m01s15i202": ("meridional_wind_at_pressure_levels", "face"),
    "m01s15i242": ("vertical_wind_at_pressure_levels", "face"),
    "m01s16i004": ("air_temperature", "face"),
    "m01s16i162": (
        "tendency_of_atmosphere_water_vapor_content_due_to_pc2_initiation",
        "face",
    ),
    "m01s16i163": (
        "tendency_of_atmosphere_cloud_liquid_water_content_due_to_pc2_initiation",
        "face",
    ),
    "m01s16i164": (
        "tendency_of_atmosphere_cloud_ice_water_content_due_to_pc2_initiation",
        "face",
    ),
    "m01s16i172": (
        "tendency_of_cloud_amount_in_atmosphere_layer_due_to_pc2_initiation",
        "face",
    ),
    "m01s16i173": (
        "tendency_of_liquid_cloud_amount_in_atmosphere_layer_due_to_pc2_initiation",
        "face",
    ),
    "m01s16i174": (
        "tendency_of_frozen_cloud_amount_in_atmosphere_layer_due_to_pc2_initiation",
        "face",
    ),
    "m01s16i182": (
        "tendency_of_atmosphere_water_vapor_content_due_to_pc2_pressure_change",
        "face",
    ),
    "m01s16i183": (
        "tendency_of_atmosphere_cloud_liquid_water_content_due_to_pc2_pressure_change",
        "face",
    ),
    "m01s16i192": (
        "tendency_of_cloud_amount_in_atmosphere_layer_due_to_pc2_pressure_change",
        "face",
    ),
    "m01s16i193": (
        "tendency_of_liquid_cloud_amount_in_atmosphere_layer_due_to_pc2_pressure_change",
        "face",
    ),
    "m01s16i202": ("geopotential_height_at_pressure_levels", "face"),
    "m01s16i203": ("temperature_at_pressure_levels", "face"),
    "m01s16i204": ("relative_humidity_wrt_ice_at_pressure_levels", "face"),
    "m01s16i206": ("specific_cloud_condensate", "face"),
    "m01s16i207": ("specific_total_water", "face"),
    "m01s16i222": ("air_pressure_at_mean_sea_level", "face"),
    "m01s16i256": ("relative_humidity_wrt_water_at_pressure_levels", "face"),
    "m01s30i112": ("wbig_eq_1_where_wphysics_gt_1", "face"),
    "m01s30i113": ("relative_humidity_over_ice_below_freezing", "face"),
    "m01s30i185": ("total_eastward_wind_increment", "face"),
    "m01s30i186": ("total_northward_wind_increment", "face"),
    "m01s30i187": ("total_vertical_air_velocity_increment", "face"),
    "m01s30i195": ("total_vapour_increment", "face"),
    "m01s30i196": ("total_liquid_water_increment", "face"),
    "m01s30i197": ("total_frozen_water_increment", "face"),
    "m01s30i201": ("zonal_wind_at_pressure_levels_for_climate_averaging", "face"),
    "m01s30i202": ("meridional_wind_at_pressure_levels_for_climate_averaging", "face"),
    "m01s30i203": ("vertical_wind_at_pressure_levels_for_climate_averaging", "face"),
    "m01s30i204": ("temperature_at_pressure_levels_for_climate_averaging", "face"),
    "m01s30i205": (
        "vapour_specific_humidity_at_pressure_levels_for_climate_averaging",
        "face",
    ),
    "m01s30i206": (
        "relative_humidity_wrt_ice_at_pressure_levels_for_climate_averaging",
        "face",
    ),
    "m01s30i207": (
        "geopotential_height_at_pressure_levels_for_climate_averaging",
        "face",
    ),
    "m01s30i208": ("omega_at_pressure_levels_for_climate_averaging", "face"),
    "m01s30i211": ("square_of_eastward_wind", "face"),
    "m01s30i212": ("product_of_eastward_wind_and_northward_wind", "face"),
    "m01s30i213": ("product_of_eastward_wind_and_upward_air_velocity", "face"),
    "m01s30i214": ("product_of_eastward_wind_and_air_temperature", "face"),
    "m01s30i215": ("product_of_eastward_wind_and_specific_humidity", "face"),
    "m01s30i217": ("product_of_eastward_wind_and_geopotential_height", "face"),
    "m01s30i218": (
        "product_of_eastward_wind_and_lagrangian_tendency_of_air_pressure",
        "face",
    ),
    "m01s30i222": ("square_of_northward_wind", "face"),
    "m01s30i223": ("product_of_northward_wind_and_upward_air_velocity", "face"),
    "m01s30i224": ("product_of_northward_wind_and_air_temperature", "face"),
    "m01s30i225": ("product_of_northward_wind_and_specific_humidity", "face"),
    "m01s30i227": ("product_of_northward_wind_and_geopotential_height", "face"),
    "m01s30i228": (
        "product_of_northward_wind_and_lagrangian_tendency_of_air_pressure",
        "face",
    ),
    "m01s30i233": ("square_of_upward_air_velocity", "face"),
    "m01s30i234": ("product_of_upward_air_velocity_and_air_temperature", "face"),
    "m01s30i235": ("product_of_upward_air_velocity_and_specific_humidity", "face"),
    "m01s30i237": ("product_of_upward_air_velocity_and_geopotential_height", "face"),
    "m01s30i238": (
        "product_of_upward_air_velocity_and_lagrangian_tendency_of_air_pressure",
        "face",
    ),
    "m01s30i244": ("square_of_air_temperature", "face"),
    "m01s30i245": ("product_of_air_temperature_and_specific_humidity", "face"),
    "m01s30i247": ("product_of_air_temperature_and_geopotential_height", "face"),
    "m01s30i248": (
        "product_of_lagrangian_tendency_of_air_pressure_and_air_temperature",
        "face",
    ),
    "m01s30i255": ("square_of_specific_humidity", "face"),
    "m01s30i257": ("product_of_specific_humidity_and_geopotential_height", "face"),
    "m01s30i258": (
        "product_of_lagrangian_tendency_of_air_pressure_and_specific_humidity",
        "face",
    ),
    "m01s30i277": ("square_of_geopotential_height", "face"),
    "m01s30i278": (
        "product_of_lagrangian_tendency_of_air_pressure_and_geopotential_height",
        "face",
    ),
    "m01s30i288": ("square_of_lagrangian_tendency_of_air_pressure", "face"),
    "m01s30i301": ("heaviside_function_at_pressure_levels", "face"),
    "m01s30i402": ("atmosphere_kinetic_energy_content", "face"),
    "m01s30i403": ("atmosphere_mass_of_air_per_unit_area", "face"),
    "m01s30i404": ("atmosphere_wetplusdry_mass_per_unit_area", "face"),
    "m01s30i405": ("atmosphere_mass_content_of_cloud_liquid_water", "face"),
    "m01s30i406": ("atmosphere_mass_content_of_cloud_ice", "face"),
    "m01s30i419": ("energy_correction_rate", "face"),
    "m01s30i420": ("atmosphere_energy_content", "face"),
    "m01s30i421": ("atmosphere_potential_energy_content", "face"),
    "m01s30i455": ("vertical_vorticity_at_pressure_levels", "face"),
    "m01s30i461": ("atmosphere_mass_content_of_water_vapor", "face"),
    "m01s34i072": ("mass_fraction_of_sulfur_dioxide_in_air", "face"),
    "m01s34i101": ("nucleation_soluble_mode_number", "face"),
    "m01s34i102": ("nucleation_soluble_mode_h2so4_mmr", "face"),
    "m01s34i103": ("aitken_soluble_mode_number", "face"),
    "m01s34i104": ("aitken_soluble_mode_h2so4_mmr", "face"),
    "m01s34i105": ("aitken_soluble_mode_black_carbon_mmr", "face"),
    "m01s34i106": ("aitken_soluble_mode_organic_matter_mmr", "face"),
    "m01s34i107": ("accumulation_soluble_mode_number", "face"),
    "m01s34i108": ("accumulation_soluble_mode_h2so4_mmr", "face"),
    "m01s34i109": ("accumulation_soluble_mode_black_carbon_mmr", "face"),
    "m01s34i110": ("accumulation_soluble_mode_organic_matter_mmr", "face"),
    "m01s34i111": ("accumulation_soluble_mode_sea_salt_mmr", "face"),
    "m01s34i112": ("accumulation_soluble_mode_dust_mmr", "face"),
    "m01s34i113": ("coarse_soluble_mode_number", "face"),
    "m01s34i114": ("coarse_soluble_mode_h2so4_mmr", "face"),
    "m01s34i115": ("coarse_soluble_mode_black_carbon_mmr", "face"),
    "m01s34i116": ("coarse_soluble_mode_organic_matter_mmr", "face"),
    "m01s34i117": ("coarse_soluble_mode_sea_salt_mmr", "face"),
    "m01s34i118": ("coarse_soluble_mode_dust_mmr", "face"),
    "m01s34i119": ("aitken_insoluble_mode_number", "face"),
    "m01s34i120": ("aitken_insoluble_mode_black_carbon_mmr", "face"),
    "m01s34i121": ("aitken_insoluble_mode_organic_matter_mmr", "face"),
    "m01s34i122": ("accumulation_insoluble_mode_number", "face"),
    "m01s34i123": ("accumulation_insoluble_mode_dust_mmr", "face"),
    "m01s34i124": ("coarse_insoluble_mode_number", "face"),
    "m01s34i125": ("coarse_insoluble_mode_dust_mmr", "face"),
    "m01s34i126": ("nucleation_soluble_mode_organic_matter_mmr", "face"),
    "m01s34i150": ("ageofair", "face"),
    "m01s35i024": ("potential_temperature_increment_from_spt", "face"),
    "m01s35i025": ("vapour_increment_from_spt", "face"),
}

# Local additions to make CSET work with RNS for RAL3-LFRic.
# TODO: Port to Paul's FCM branch.
RAL3_LFRic_mods = {
    "m01s03i281": ("visibility_in_air", "face"),  # inc precipitation.
    "m01s01i235": ("surface_downward_shortwave_flux", "face"),
    "m01s02i207": ("surface_downward_longwave_flux", "face"),
    "m01s03i250": ("dew_point_temperature_at_screen_level", "face"),
    "m01s03i025": ("atmosphere_boundary_layer_thickness", "face"),
    "m01s09i203": ("low_type_cloud_area_fraction", "face"),
    "m01s09i204": ("medium_type_cloud_area_fraction", "face"),
    "m01s09i205": ("high_type_cloud_area_fraction", "face"),
    "m01s09i210": ("cloud_base_altitude", "face"),
    "m01s04i112": ("radar_reflectivity_at_1km_above_the_surface", "face"),
    "m01s00i391": ("humidity_mixing_ratio", "face"),
    "m01s00i255": ("exner_pressure_at_cell_interfaces", "face"),
    "m01s09i231": ("combined_cloud_amount", "face"),
    "m01s21i104": ("number_of_lightning_flashes_in_column", "face"),
    "m01s15i229": ("potential_vorticity_at_pressure_levels", "face"),
    "m01s16i205": ("wet_bulb_potential_temperature_at_pressure_levels", "face"),
}


# Combine into single STASH dict
STASH_TO_LFRIC.update(RAL3_LFRic_mods)



"""Configuration file for the Sphinx documentation builder.

For the full list of built-in configuration values, see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "CSET"
copyright = "Crown copyright, Met Office (2022-2024) and CSET contributors."
author = "Met Office and Partners"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_baseurl = "https://metoffice.github.io/CSET/"

# -- LaTeX output configuration ----------------------------------------------
# https://www.sphinx-doc.org/en/master/latex.html

latex_engine = "xelatex"
latex_elements = {"papersize": "a4paper"}
latex_show_urls = "footnote"

# -- Extlinks configuration --------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/extlinks.html

extlinks = {
    "issue": ("https://github.com/MetOffice/CSET/issues/%s", "Issue #%s"),
    "pr": ("https://github.com/MetOffice/CSET/pull/%s", "PR #%s"),
}

# Ignore GitHub links as we have many and don't want to be rate limited.
linkcheck_ignore = [
    r"https://github.com/.+/(issues|pull)/\d+",
]



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests regrid operator."""

import iris
import iris.coord_systems
import iris.cube
import numpy as np
import pytest

from CSET.operators import _utils, read, regrid


# Session scope fixtures, so the test data only has to be loaded once.
@pytest.fixture(scope="session")
def regrid_source_cube() -> iris.cube.Cube:
    """Get a cube to regrid."""
    return read.read_cube(
        "tests/test_data/regrid/regrid_rectilinearGeogCS.nc", "surface_altitude"
    )


@pytest.fixture(scope="session")
def regrid_test_cube() -> iris.cube.Cube:
    """Get a regridded cube to compare."""
    return read.read_cube("tests/test_data/regrid/out_rectilinearGeogCS_0p5deg.nc")


def test_regrid_onto_cube(regrid_source_cube, regrid_test_cube):
    """Test regrid case where target cube to project onto is specified."""
    # Test 1: Rectilinear GeogCS grid case
    assert np.allclose(
        regrid.regrid_onto_cube(
            regrid_source_cube, regrid_test_cube, method="Linear"
        ).data,
        regrid_test_cube.data,
        rtol=1e-06,
        atol=1e-02,
    )


def test_regrid_onto_cube_cubes(regrid_source_cube, regrid_test_cube):
    """Test regrid case where target cube to project onto is specified for multiple cubes."""
    # Create cubelist with multiple cubes.
    cubelist_to_regrid = iris.cube.CubeList([regrid_source_cube, regrid_source_cube])
    # Regrid all those cubes.
    regridded_cubes = regrid.regrid_onto_cube(
        cubelist_to_regrid, regrid_test_cube, method="Linear"
    )
    # Check it regridded correctly.
    assert len(regridded_cubes) == 2
    for cube in regridded_cubes:
        lat_name, lon_name = _utils.get_cube_yxcoordname(cube)
        assert (
            cube.coord(lat_name).points == regrid_test_cube.coord(lat_name).points
        ).all()
        assert (
            cube.coord(lon_name).points == regrid_test_cube.coord(lon_name).points
        ).all()


def test_regrid_onto_xyspacing_cubes(regrid_source_cube):
    """Test regrid case where xyspacing to project onto is specified for multiple cubes."""
    # Create cubelist with multiple cubes
    cubelist_to_regrid = iris.cube.CubeList([regrid_source_cube, regrid_source_cube])
    regridded_cubes = regrid.regrid_onto_xyspacing(
        cubelist_to_regrid, xspacing=0.5, yspacing=0.5, method="Linear"
    )
    # Check it regridded correctly.
    assert len(regridded_cubes) == 2
    for cube in regridded_cubes:
        lat_name, lon_name = _utils.get_cube_yxcoordname(cube)
        lat_coord = cube.coord(lat_name)
        lon_coord = cube.coord(lon_name)
        # Inclusive interval from -1.0 to 6.5 in steps of 0.5.
        assert (lat_coord.points == [x / 10.0 for x in range(-10, 70, 5)]).all()
        # Inclusive interval from 32.0 to 39.5 in steps of 0.5.
        assert (lon_coord.points == [x / 10.0 for x in range(320, 400, 5)]).all()


def test_regrid_onto_cube_unknown_crs(regrid_source_cube, regrid_test_cube):
    """Coordinate reference system is unrecognised."""
    lat_name, lon_name = _utils.get_cube_yxcoordname(regrid_source_cube)
    # Exchange X to unsupported coordinate system.
    source_changed_x = regrid_source_cube.copy()
    source_changed_x.coord(lon_name).coord_system = iris.coord_systems.OSGB()
    with pytest.raises(NotImplementedError):
        regrid.regrid_onto_cube(source_changed_x, regrid_test_cube, method="Linear")

    # Exchange Y to unsupported coordinate system.
    source_changed_y = regrid_source_cube.copy()
    source_changed_y.coord(lat_name).coord_system = iris.coord_systems.OSGB()
    with pytest.raises(NotImplementedError):
        regrid.regrid_onto_cube(source_changed_y, regrid_test_cube, method="Linear")


def test_regrid_onto_cube_non_callable_method(regrid_source_cube, regrid_test_cube):
    """Method exists but is not callable."""
    with pytest.raises(NotImplementedError):
        regrid.regrid_onto_cube(regrid_source_cube, regrid_test_cube, method="maths")


def test_regrid_onto_cube_unknown_method(regrid_source_cube, regrid_test_cube):
    """Method does not exist."""
    with pytest.raises(NotImplementedError):
        regrid.regrid_onto_cube(
            regrid_source_cube, regrid_test_cube, method="nonexistent"
        )


def test_regrid_onto_xyspacing(regrid_source_cube, regrid_test_cube):
    """Test regrid case where x and y spacing are specified."""
    # Test 1: Rectilinear GeogCS grid case
    assert np.allclose(
        regrid.regrid_onto_xyspacing(
            regrid_source_cube, xspacing=0.5, yspacing=0.5, method="Linear"
        ).data,
        regrid_test_cube.data,
        rtol=1e-06,
        atol=1e-02,
    )


def test_regrid_onto_xyspacing_unknown_crs(regrid_source_cube):
    """Coordinate reference system is unrecognised."""
    lat_name, lon_name = _utils.get_cube_yxcoordname(regrid_source_cube)
    # Exchange X to unsupported coordinate system.
    source_changed_x = regrid_source_cube.copy()
    source_changed_x.coord(lon_name).coord_system = iris.coord_systems.OSGB()
    with pytest.raises(NotImplementedError):
        regrid.regrid_onto_xyspacing(
            source_changed_x, xspacing=0.5, yspacing=0.5, method="Linear"
        )

    # Exchange Y to unsupported coordinate system.
    source_changed_y = regrid_source_cube.copy()
    source_changed_y.coord(lat_name).coord_system = iris.coord_systems.OSGB()
    with pytest.raises(NotImplementedError):
        regrid.regrid_onto_xyspacing(
            source_changed_y, xspacing=0.5, yspacing=0.5, method="Linear"
        )


def test_regrid_onto_xyspacing_non_callable_method(regrid_source_cube):
    """Method exists but is not callable."""
    with pytest.raises(NotImplementedError):
        regrid.regrid_onto_xyspacing(
            regrid_source_cube, xspacing=0.5, yspacing=0.5, method="maths"
        )


def test_regrid_onto_xyspacing_unknown_method(regrid_source_cube):
    """Method does not exist."""
    with pytest.raises(NotImplementedError):
        regrid.regrid_onto_xyspacing(
            regrid_source_cube, xspacing=0.5, yspacing=0.5, method="nonexistent"
        )


def test_regrid_to_single_point_east(cube):
    """Test extracting a single point.

    Test that the grid latitude rotation works when the
    centre of the grid is too far east (that is, the centre
    of the longitude range is further east than 180 deg).
    """
    cube_fix = read._longitude_fix_callback(cube, None, None)
    regrid_cube = regrid.regrid_to_single_point(
        cube_fix, 0.5, -1.5, "rotated", "Nearest", boundary_margin=1
    )
    lat_name, lon_name = _utils.get_cube_yxcoordname(regrid_cube)
    lat_coord = regrid_cube.coord(lat_name)
    lon_coord = regrid_cube.coord(lon_name)
    assert lat_coord.shape == (1,)
    assert lon_coord.shape == (1,)
    assert lat_coord.points[0] == 0.5
    assert lon_coord.points[0] == -1.5


def test_regrid_to_single_point_west(cube):
    """Test extracting a single point.

    Test that the grid latitude rotation works when the
    centre of the grid is too far west (that is, the centre
    of the longitude range is further west than -180 deg).
    """
    long_coord = cube.coord("grid_longitude").points.copy()
    long_coord -= 720.0
    cube.coord("grid_longitude").points = long_coord
    cube_fix = read._longitude_fix_callback(cube, None, None)
    regrid_cube = regrid.regrid_to_single_point(
        cube_fix, 0.5, -1.5, "rotated", "Nearest", boundary_margin=1
    )
    lat_name, lon_name = _utils.get_cube_yxcoordname(regrid_cube)
    lat_coord = regrid_cube.coord(lat_name)
    lon_coord = regrid_cube.coord(lon_name)
    assert lat_coord.shape == (1,)
    assert lon_coord.shape == (1,)
    assert lat_coord.points[0] == 0.5
    assert lon_coord.points[0] == -1.5


def test_regrid_to_single_point_longitude_transform_1(cube):
    """Test extracting a single point.

    Test that, if a longitude selection is made that is too
    far east (further east than 180 deg), the code corrects it
    back into the standard range (-180 deg to 180 deg).
    """
    regrid_cube = regrid.regrid_to_single_point(
        cube, 0.5, 358.5, "rotated", "Nearest", boundary_margin=1
    )
    lat_name, lon_name = _utils.get_cube_yxcoordname(regrid_cube)
    lat_coord = regrid_cube.coord(lat_name)
    lon_coord = regrid_cube.coord(lon_name)
    assert lat_coord.shape == (1,)
    assert lon_coord.shape == (1,)
    assert lat_coord.points[0] == 0.5
    assert lon_coord.points[0] == -1.5


def test_regrid_to_single_point_longitude_transform_2(cube):
    """Test extracting a single point.

    Test that, if a longitude selection is made that is too
    far west (further west than -180 deg), the code corrects it
    back into the standard range (-180 deg to 180 deg).
    """
    regrid_cube = regrid.regrid_to_single_point(
        cube, 0.5, -361.5, "rotated", "Nearest", boundary_margin=1
    )
    lat_name, lon_name = _utils.get_cube_yxcoordname(regrid_cube)
    lat_coord = regrid_cube.coord(lat_name)
    lon_coord = regrid_cube.coord(lon_name)
    assert lat_coord.shape == (1,)
    assert lon_coord.shape == (1,)
    assert lat_coord.points[0] == 0.5
    assert lon_coord.points[0] == -1.5


def test_regrid_to_single_point_realworld(cube):
    """Test extracting a single point.

    Test that, if a real world coordinate is specified, the code is
    mapping this point correctly onto the rotated grid.
    """
    cube_fix = read._longitude_fix_callback(cube, None, None)
    regrid_cube = regrid.regrid_to_single_point(
        cube_fix, 52.98, -5.0, "realworld", "Nearest", boundary_margin=1
    )
    expected_array = "array(288.59375, dtype=float32)"
    assert repr(regrid_cube[0].data) == expected_array


def test_regrid_to_single_point_rotated(cube):
    """Test extracting a single point.

    Test that, if a rotated coordinate is specified, the answer
    matches the corresponding coordinate on the real world grid.
    """
    cube_fix = read._longitude_fix_callback(cube, None, None)
    regrid_cube = regrid.regrid_to_single_point(
        cube_fix, 0.5, -1.5, "rotated", "Nearest", boundary_margin=1
    )
    expected_array = "array(288.59375, dtype=float32)"
    assert repr(regrid_cube[0].data) == expected_array


def test_regrid_to_single_point_missing_coord(cube):
    """Missing coordinate raises error."""
    # Missing X coordinate.
    source = cube.copy()
    source.remove_coord("grid_longitude")
    with pytest.raises(ValueError):
        regrid.regrid_to_single_point(
            source, 0.5, 358.5, "rotated", "Nearest", boundary_margin=1
        )

    # Missing Y coordinate.
    source = cube.copy()
    source.remove_coord("grid_latitude")
    with pytest.raises(ValueError):
        regrid.regrid_to_single_point(source, 0.5, 358.5, "rotated", "Nearest")


def test_longitude_fix_callback_missing_coord(cube):
    """Missing coordinate raises error."""
    # Missing X coordinate.
    source = cube.copy()
    source.remove_coord("grid_longitude")
    cube_fix = read._longitude_fix_callback(source.copy(), None, None)
    assert cube_fix == source

    # Missing Y coordinate.
    source = cube.copy()
    source.remove_coord("grid_latitude")
    cube_fix = read._longitude_fix_callback(source.copy(), None, None)
    assert cube_fix == source


def test_regrid_to_single_point_unknown_crs_x(cube):
    """X coordinate reference system is unrecognised."""
    # Exchange to unsupported coordinate system.
    cube.coord("grid_longitude").coord_system = iris.coord_systems.OSGB()
    with pytest.raises(NotImplementedError):
        regrid.regrid_to_single_point(cube, 0.5, 358.5, "rotated", "Nearest")


def test_regrid_to_single_point_unknown_crs_y(cube):
    """Y coordinate reference system is unrecognised."""
    # Exchange to unsupported coordinate system.
    cube.coord("grid_latitude").coord_system = iris.coord_systems.OSGB()
    with pytest.raises(NotImplementedError):
        regrid.regrid_to_single_point(cube, 0.5, 358.5, "rotated", "Nearest")


def test_regrid_to_single_point_outside_domain_longitude(regrid_source_cube):
    """Error if coordinates are outside the model domain."""
    with pytest.raises(ValueError):
        regrid.regrid_to_single_point(
            regrid_source_cube, 0.5, 178.5, "rotated", "Nearest"
        )


def test_regrid_to_single_point_outside_domain_latitude(regrid_source_cube):
    """Error if coordinates are outside the model domain."""
    with pytest.raises(ValueError):
        regrid.regrid_to_single_point(
            regrid_source_cube, 80.5, 358.5, "rotated", "Nearest"
        )


@pytest.mark.filterwarnings("ignore:Selected point is within")
def test_regrid_to_single_point_unknown_method(cube):
    """Method does not exist."""
    with pytest.raises(NotImplementedError):
        regrid.regrid_to_single_point(cube, 0.5, 358.5, "rotated", method="nonexistent")


@pytest.mark.filterwarnings(
    "ignore:this date/calendar/year zero convention is not supported by CF"
)
def test_boundary_warning(regrid_source_cube):
    """Ensures a warning is raised when chosen point is too close to boundary."""
    with pytest.warns(
        regrid.BoundaryWarning, match="Selected point is within 8 gridlengths"
    ) as warning:
        regrid.regrid_to_single_point(
            regrid_source_cube, -0.9, 393.0, "rotated", "Nearest"
        )

    assert len(warning) == 1
    assert issubclass(warning[-1].category, regrid.BoundaryWarning)



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Miscellaneous operators."""

import itertools
import logging
from collections.abc import Iterable

import iris
import iris.coords
import numpy as np
from iris.cube import Cube, CubeList

from CSET._common import iter_maybe
from CSET.operators._utils import (
    fully_equalise_attributes,
    get_cube_yxcoordname,
)
from CSET.operators.regrid import regrid_onto_cube


def noop(x, **kwargs):
    """Return its input without doing anything to it.

    Useful for constructing diagnostic chains.

    Arguments
    ---------
    x: Any
        Input to return.

    Returns
    -------
    x: Any
        The input that was given.
    """
    return x


def remove_attribute(
    cubes: Cube | CubeList, attribute: str | Iterable, **kwargs
) -> CubeList:
    """Remove a cube attribute.

    If the attribute is not on the cube, the cube is passed through unchanged.

    Arguments
    ---------
    cubes: Cube | CubeList
        One or more cubes to remove the attribute from.
    attribute: str | Iterable
        Name of attribute (or Iterable of names) to remove.

    Returns
    -------
    cubes: CubeList
        CubeList of cube(s) with the attribute removed.
    """
    # Ensure cubes is a CubeList.
    if not isinstance(cubes, CubeList):
        cubes = CubeList(iter_maybe(cubes))

    for cube in cubes:
        for attr in iter_maybe(attribute):
            cube.attributes.pop(attr, None)
    return cubes


def addition(addend_1, addend_2):
    """Addition of two fields.

    Parameters
    ----------
    addend_1: Cube
        Any field to have another field added to it.
    addend_2: Cube
        Any field to be added to another field.

    Returns
    -------
    Cube

    Raises
    ------
    ValueError, iris.exceptions.NotYetImplementedError
        When the cubes are not compatible.

    Notes
    -----
    This is a simple operator designed for combination of diagnostics or
    creating new diagnostics by using recipes.

    Examples
    --------
    >>> field_addition = misc.addition(kinetic_energy_u, kinetic_energy_v)

    """
    return addend_1 + addend_2


def subtraction(minuend, subtrahend):
    """Subtraction of two fields.

    Parameters
    ----------
    minuend: Cube
        Any field to have another field subtracted from it.
    subtrahend: Cube
        Any field to be subtracted from to another field.

    Returns
    -------
    Cube

    Raises
    ------
    ValueError, iris.exceptions.NotYetImplementedError
        When the cubes are not compatible.

    Notes
    -----
    This is a simple operator designed for combination of diagnostics or
    creating new diagnostics by using recipes. It can be used for model
    differences to allow for comparisons between the same field in different
    models or model configurations.

    Examples
    --------
    >>> model_diff = misc.subtraction(temperature_model_A, temperature_model_B)

    """
    return minuend - subtrahend


def division(numerator, denominator):
    """Division of two fields.

    Parameters
    ----------
    numerator: Cube
        Any field to have the ratio taken with respect to another field.
    denominator: Cube
        Any field used to divide another field or provide the reference
        value in a ratio.

    Returns
    -------
    Cube

    Raises
    ------
    ValueError
        When the cubes are not compatible.

    Notes
    -----
    This is a simple operator designed for combination of diagnostics or
    creating new diagnostics by using recipes.

    Examples
    --------
    >>> bowen_ratio = misc.division(sensible_heat_flux, latent_heat_flux)

    """
    return numerator / denominator


def multiplication(multiplicand, multiplier):
    """Multiplication of two fields.

    Parameters
    ----------
    multiplicand: Cube
        Any field to be multiplied by another field.
    multiplier: Cube
        Any field to be multiplied to another field.

    Returns
    -------
    Cube

    Raises
    ------
    ValueError
        When the cubes are not compatible.

    Notes
    -----
    This is a simple operator designed for combination of diagnostics or
    creating new diagnostics by using recipes.

    Examples
    --------
    >>> filtered_CAPE_ratio = misc.multiplication(CAPE_ratio, inflow_layer_properties)

    """
    return multiplicand * multiplier


def combine_cubes_into_cubelist(first: Cube | CubeList, **kwargs) -> CubeList:
    """Operator that combines multiple cubes or CubeLists into one.

    Arguments
    ---------
    first: Cube | CubeList
        First cube or CubeList to merge into CubeList.
    second: Cube | CubeList
        Second cube or CubeList to merge into CubeList. This must be a named
        argument.
    third: Cube | CubeList
        There can be any number of additional arguments, they just need unique
        names.
    ...

    Returns
    -------
    combined_cubelist: CubeList
        Combined CubeList containing all cubes/CubeLists.

    Raises
    ------
    TypeError:
        If the provided arguments are not either a Cube or CubeList.
    """
    # Create empty CubeList to store cubes/CubeList.
    all_cubes = CubeList()
    # Combine all CubeLists into a single flat iterable.
    for item in itertools.chain(iter_maybe(first), *map(iter_maybe, kwargs.values())):
        # Check each item is a Cube, erroring if not.
        if isinstance(item, Cube):
            # Add cube to CubeList.
            all_cubes.append(item)
        else:
            raise TypeError("Not a Cube or CubeList!", item)
    return all_cubes


def difference(cubes: CubeList):
    """Difference of two fields.

    Parameters
    ----------
    cubes: CubeList
        A list of exactly two cubes. One must have the cset_comparison_base
        attribute set to 1, and will be used as the base of the comparison.

    Returns
    -------
    Cube

    Raises
    ------
    ValueError
        When the cubes are not compatible.

    Notes
    -----
    This is a simple operator designed for combination of diagnostics or
    creating new diagnostics by using recipes. It can be used for model
    differences to allow for comparisons between the same field in different
    models or model configurations.

    Examples
    --------
    >>> model_diff = misc.difference(temperature_model_A, temperature_model_B)

    """
    if len(cubes) != 2:
        raise ValueError("cubes should contain exactly 2 cubes.")
    base = cubes.extract_cube(iris.AttributeConstraint(cset_comparison_base=1))
    other = cubes.extract_cube(
        iris.Constraint(
            cube_func=lambda cube: "cset_comparison_base" not in cube.attributes
        )
    )

    # Get spatial coord names.
    base_lat_name, base_lon_name = get_cube_yxcoordname(base)
    other_lat_name, other_lon_name = get_cube_yxcoordname(other)

    # Check latitude, longitude shape the same. Not comparing points, as these
    # might slightly differ due to rounding errors (especially in future if we
    # are regridding cubes to common resolutions).
    # An exception has been included here to deal with some variables on different
    # grid staggering (cell center vs edge of cell) when comparing UM to LFRic,
    # depending on whether they are on a B grid or Arakawa staggering. Note we do not
    # generally apply regridding to any variable where dimension sizes do not
    # match, to make sure we are using appropriate regridding technique. In this
    # case for winds, a Linear regridding is appropriate (smooth variable).
    if (
        base.coord(base_lat_name).shape != other.coord(other_lat_name).shape
        or base.coord(base_lon_name).shape != other.coord(other_lon_name).shape
    ):
        if base.attributes.get("STASH", None) in [
            "m01s03i225",
            "m01s03i226",
            "m01s15i201",
            "m01s15i202",
            "m01s15i229",
            "m01s30i205",
        ]:
            base = regrid_onto_cube(base, other, method="Linear")
        else:
            raise ValueError(
                f"Cubes should have the same shape, got {base.coord(base_lat_name)} {other.coord(other_lat_name)}"
            )

    def is_increasing(sequence: list) -> bool:
        """Determine the direction of an ordered sequence.

        Returns "increasing" or "decreasing" depending on whether the sequence
        is going up or down. The sequence should already be monotonic, with no
        duplicate values. An iris DimCoord's points fulfills this criteria.
        """
        return sequence[0] < sequence[1]

    # Figure out if we are comparing between UM and LFRic; flip array if so.
    base_lat_direction = is_increasing(base.coord(base_lat_name).points)
    other_lat_direction = is_increasing(other.coord(other_lat_name).points)
    if base_lat_direction != other_lat_direction:
        other.data = np.flip(other.data, other.coord(other_lat_name).cube_dims(other))

    # Extract just common time points.
    logging.debug("Base: %s\nOther: %s", base.coord("time"), other.coord("time"))
    base_times = set(base.coord("time").units.num2date(base.coord("time").points))
    other_times = set(other.coord("time").units.num2date(other.coord("time").points))
    shared_times = set.intersection(base_times, other_times)
    time_constraint = iris.Constraint(time=lambda cell: cell.point in shared_times)
    base = base.extract(time_constraint)
    other = other.extract(time_constraint)

    # Equalise attributes so we can merge.
    fully_equalise_attributes([base, other])
    logging.debug("Base: %s\nOther: %s", base, other)

    # This currently relies on the cubes having the same underlying data layout.
    difference = base.copy()
    difference.rename(base.name() + "_difference")
    difference.data = base.data - other.data
    return difference



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the command line interface.

In many ways these are integration tests.
"""

import logging
import subprocess
from pathlib import Path
from uuid import uuid4

import pytest

import CSET
import CSET.operators


def test_command_line_invocation():
    """Check invocation via different entrypoints.

    Uses subprocess to validate the external interface.
    """
    # Invoke via cset command
    subprocess.run(["cset", "--version"], check=True)
    # Invoke via __main__.py
    subprocess.run(["python3", "-m", "CSET", "--version"], check=True)


# Every other test should not use the command line interface, but rather stay
# within python to ensure coverage measurement.
def test_argument_parser():
    """Tests the argument parser behaves appropriately."""
    parser = CSET.setup_argument_parser()
    # Test verbose flag.
    args = parser.parse_args(["recipe-id", "-r", "recipe.yaml"])
    assert args.verbose == 0
    args = parser.parse_args(["-v", "recipe-id", "-r", "recipe.yaml"])
    assert args.verbose == 1
    args = parser.parse_args(["-vv", "recipe-id", "-r", "recipe.yaml"])
    assert args.verbose == 2


def test_setup_logging():
    """Tests the logging setup at various verbosity levels."""
    root_logger = logging.getLogger()
    # Log level gets pinned at a minimum of INFO for file output.
    CSET.setup_logging(0)
    assert root_logger.level == logging.INFO
    # -v
    CSET.setup_logging(1)
    assert root_logger.level == logging.INFO
    # -vv
    CSET.setup_logging(2)
    assert root_logger.level == logging.DEBUG


def test_main_no_subparser(capsys):
    """Appropriate error when no subparser is given."""
    with pytest.raises(SystemExit) as sysexit:
        CSET.main(["cset"])
    assert sysexit.value.code == 127
    assert capsys.readouterr().err == "Please choose a command.\n"


def test_main_unhandled_exception_normal(capsys):
    """Appropriate error when an unhandled exception occurs."""
    with pytest.raises(SystemExit) as sysexit:
        CSET.main(
            [
                "cset",
                "bake",
                "--recipe=/non-existent/recipe.yaml",
                "--input-dir=/dev/null",
                "--output-dir=/dev/null",
            ]
        )
    assert sysexit.value.code == 1
    assert (
        capsys.readouterr().err
        == "[Errno 2] No such file or directory: '/non-existent/recipe.yaml'\n"
    )


def test_main_unhandled_exception_debug(caplog, capsys):
    """Appropriate error when an unhandled exception occurs under debug mode."""
    with pytest.raises(FileNotFoundError):
        CSET.main(
            [
                "cset",
                "-vv",
                "bake",
                "--recipe=/non-existent/recipe.yaml",
                "--input-dir=/dev/null",
                "--output-dir=/dev/null",
            ]
        )
    assert (
        "[Errno 2] No such file or directory: '/non-existent/recipe.yaml'\n"
        in capsys.readouterr().err
    )
    log_record = caplog.records[-1]
    assert log_record.message == "An unhandled exception occurred."
    assert log_record.levelno == logging.DEBUG


def test_bake_recipe_execution(monkeypatch):
    """Test running CSET recipe from the command line."""
    bake_ran = False

    def _bake_test(args, unparsed_args):
        nonlocal bake_ran
        bake_ran = True
        assert args.input_dir == ["/dev/null"]
        assert args.output_dir == Path("/dev/null")
        assert args.recipe == Path("tests/test_data/noop_recipe.yaml")

    monkeypatch.setattr(CSET, "_bake_command", _bake_test)
    CSET.main(
        [
            "cset",
            "bake",
            "--input-dir=/dev/null",
            "--output-dir=/dev/null",
            "--recipe=tests/test_data/noop_recipe.yaml",
        ]
    )


def test_bake_invalid_args(capsys):
    """Invalid arguments give non-zero exit code."""
    with pytest.raises(SystemExit) as sysexit:
        CSET.main(
            [
                "cset",
                "bake",
                "--recipe=foo",
                "--input-dir=/tmp",
                "--output-dir=/tmp",
                "--not-a-real-option",
            ]
        )
    assert sysexit.value.code == 127
    assert capsys.readouterr().err == "Unknown argument: --not-a-real-option\n"


def test_bake_invalid_args_input_dir(capsys):
    """Missing required input-dir argument for bake."""
    with pytest.raises(SystemExit) as sysexit:
        CSET.main(["cset", "bake", "--recipe=foo", "--output-dir=/tmp"])
    assert sysexit.value.code == 2
    assert capsys.readouterr().err.endswith(
        "cset bake: error: the following arguments are required: -i/--input-dir\n"
    )


def test_graph_creation(tmp_path: Path):
    """Generates a graph with the command line interface."""
    # We can't easily test running without the output specified from the CLI, as
    # the call to xdg-open breaks in CI, due to it not being a graphical
    # environment.

    # Run with output path specified
    output_file = tmp_path / f"{uuid4()}.svg"
    CSET.main(
        [
            "cset",
            "graph",
            "-o",
            str(output_file),
            "--recipe=tests/test_data/noop_recipe.yaml",
        ]
    )
    assert output_file.is_file()
    output_file.unlink()


def test_graph_details(tmp_path: Path):
    """Generate a graph with details."""
    output_file = tmp_path / f"{uuid4()}.svg"
    CSET.main(
        [
            "cset",
            "graph",
            "--details",
            "-o",
            str(output_file),
            "--recipe=tests/test_data/noop_recipe.yaml",
        ]
    )
    assert output_file.is_file()


def test_cookbook_cwd(tmp_working_dir):
    """Unpacking the recipes into the current working directory."""
    CSET.main(["cset", "cookbook", "CAPE_ratio_plot.yaml"])
    assert Path("CAPE_ratio_plot.yaml").is_file()


def test_cookbook_path(tmp_path: Path):
    """Unpacking the recipes into a specified directory."""
    CSET.main(
        ["cset", "cookbook", "--output-dir", str(tmp_path), "CAPE_ratio_plot.yaml"]
    )
    assert (tmp_path / "CAPE_ratio_plot.yaml").is_file()


def test_cookbook_list_available_recipes(capsys):
    """List all available recipes."""
    CSET.main(["cset", "cookbook", "--details"])
    stdout = capsys.readouterr().out
    # Check start.
    assert stdout.startswith("Available recipes:\n")
    # Check has some recipes.
    assert len(stdout.splitlines()) > 3


def test_cookbook_detail_recipe(capsys):
    """Show detail of a recipe."""
    CSET.main(["cset", "cookbook", "--details", "CAPE_ratio_plot.yaml"])
    assert capsys.readouterr().out.startswith("\n\tCAPE_ratio_plot.yaml\n")


def test_cookbook_non_existent_recipe(tmp_path):
    """Non-existent recipe give non-zero exit code."""
    with pytest.raises(SystemExit) as sysexit:
        CSET.main(
            ["cset", "cookbook", "--output-dir", str(tmp_path), "non-existent.yaml"]
        )
    assert sysexit.value.code == 1


def test_recipe_id(capsys):
    """Get recipe ID for a recipe."""
    CSET.main(["cset", "recipe-id", "-r", "tests/test_data/noop_recipe.yaml"])
    assert capsys.readouterr().out == "noop\n"


def test_recipe_id_no_title(capsys):
    """Get recipe id for recipe without a title."""
    CSET.main(["cset", "recipe-id", "-r", "tests/test_data/ensemble_air_temp.yaml"])
    # UUID output + newline.
    assert len(capsys.readouterr().out) == 37


def test_cset_addopts(capsys, monkeypatch):
    """Lists in CSET_ADDOPTS environment variable don't crash the parser."""
    monkeypatch.setenv("CSET_ADDOPTS", "--LIST='[1, 2, 3]'")
    CSET.main(["cset", "recipe-id", "-r", "tests/test_data/addopts_test_recipe.yaml"])
    assert capsys.readouterr().out == "list_1_2_3\n"



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Reading operator tests."""

import logging

import iris
import iris.coords
import iris.cube
import pytest

from CSET.operators import read


def test_read_cubes():
    """Read cube and verify."""
    cubes = read.read_cubes("tests/test_data/air_temp.nc")
    assert len(cubes) == 3
    expected_cubes = [
        "<iris 'Cube' of air_temperature / (K) (time: 2; grid_latitude: 17; grid_longitude: 13)>",
        "<iris 'Cube' of air_temperature / (K) (time: 3; grid_latitude: 17; grid_longitude: 13)>",
    ]
    for cube in cubes:
        assert repr(cube) in expected_cubes


def test_read_cubes_no_cubes_warning():
    """Warning emitted when constraint gives no cubes."""
    with pytest.warns(UserWarning, match="No cubes loaded"):
        cubes = read.read_cubes("tests/test_data/air_temp.nc", "non-existent")
    assert len(cubes) == 0


def test_read_cubes_ensemble_with_realization_coord():
    """Read ensemble data from a single file with a realization dimension."""
    cubes = read.read_cubes("tests/test_data/exeter_ensemble_single_file.nc")
    # Check ensemble members have been merged into a single cube.
    assert len(cubes) == 1
    cube = cubes[0]
    # Check realization is an integer.
    for point in cube.coord("realization").points:
        assert isinstance(int(point), int)


def test_read_cubes_ensemble_separate_files():
    """Read ensemble from multiple files with the member number in filename."""
    from CSET.operators import constraints

    cubes = read.read_cubes(
        "tests/test_data/",
        constraint=constraints.generate_stash_constraint("m01s03i236"),
        filename_pattern="exeter_em*.nc",
    )
    # Check ensemble members have been merged into a single cube.
    assert len(cubes) == 1
    cube = cubes[0]
    # Check realization is an integer.
    for point in cube.coord("realization").points:
        assert isinstance(int(point), int)


def test_read_cubes_verify_comparison_base():
    """Read two paths for cubes, and check only the first is marked as base."""
    paths = [
        "tests/test_data/air_temp.nc",  # UM 11.9
        "tests/test_data/exeter_em01.nc",  # UM 13.0
    ]
    cubes = read.read_cubes(paths)
    for cube in cubes.extract(iris.AttributeConstraint(um_version="11.9")):
        assert cube.attributes["cset_comparison_base"] == 1
    for cube in cubes.extract(iris.AttributeConstraint(um_version="13.0")):
        assert "cset_comparison_base" not in cube.attributes


def test_fieldsfile_ensemble_naming():
    """Extracting the realization from the fields file naming convention."""
    cube = iris.cube.Cube([0])
    filename = "myfieldsfile_enuk_um_001/enukaa_pd000"
    read._ensemble_callback(cube, None, filename)
    assert cube.coord("realization").points[0] == 1


def test_read_cube():
    """Returns a Cube rather than CubeList."""
    from CSET.operators import constraints

    cube = read.read_cube(
        "tests/test_data/air_temp.nc",
        constraint=constraints.generate_cell_methods_constraint([]),
    )
    assert isinstance(cube, iris.cube.Cube)


def test_read_cube_unconstrained():
    """Error for multiple cubes read."""
    with pytest.raises(ValueError):
        read.read_cube("tests/test_data/air_temp.nc")


def test_read_cube_merge_concatenate():
    """Cubes are combined after callbacks have been applied."""
    cube = read.read_cube("tests/test_data/concat_after_fix_[12].nc")
    assert isinstance(cube, iris.cube.Cube)


def test_check_input_files_direct_path(tmp_path):
    """Get a iterable of a single file from a direct path as a string."""
    file_path = tmp_path / "file"
    file_path.touch()
    string_path = str(file_path)
    actual = read._check_input_files(string_path, "*")
    expected = [file_path]
    assert actual == expected


def test_check_input_files_direct_path_glob(tmp_path):
    """Get a iterable of files from a direct path containing a glob pattern."""
    file1_path = tmp_path / "file1"
    file2_path = tmp_path / "file2"
    file1_path.touch()
    file2_path.touch()
    glob_path = f"{tmp_path}/file*"
    actual = read._check_input_files(glob_path, "*")
    expected = [file1_path, file2_path]
    assert actual == expected


def test_check_input_files_direct_path_match_glob_like_file(tmp_path):
    """Get a iterable of a single file from a glob-like direct path."""
    file1_path = tmp_path / "file1"
    glob_like_path = tmp_path / "file*"
    file1_path.touch()
    glob_like_path.touch()
    actual = read._check_input_files(str(glob_like_path), "*")
    expected = [glob_like_path]
    assert actual == expected


def test_check_input_files_input_directory(tmp_path):
    """Get a iterable of files in an input directory."""
    file1_path = tmp_path / "file1"
    file2_path = tmp_path / "file2"
    file1_path.touch()
    file2_path.touch()
    actual = read._check_input_files(str(tmp_path), "*")
    expected = [file1_path, file2_path]
    assert actual == expected


def test_check_input_files_input_directory_glob(tmp_path):
    """Get a iterable of files in an input directory with a glob pattern."""
    file1_path = tmp_path / "file1"
    file2_path = tmp_path / "file2"
    file1_path.touch()
    file2_path.touch()
    actual = read._check_input_files(str(tmp_path), "*1")
    expected = [file1_path]
    assert actual == expected


def test_check_input_files_invalid_path(tmp_path):
    """Error when path doesn't exist."""
    file_path = str(tmp_path / "file")
    with pytest.raises(FileNotFoundError):
        read._check_input_files(file_path, "*")


def test_check_input_files_no_file_in_directory(tmp_path):
    """Error when input directory doesn't contain any files."""
    with pytest.raises(FileNotFoundError):
        read._check_input_files(str(tmp_path), "*")


def test_um_normalise_callback_rename_stash(cube):
    """Correctly translate from STASH to LFRic variable name."""
    read._um_normalise_callback(cube, None, None)
    actual = cube.long_name
    expected = "temperature_at_screen_level"
    assert actual == expected


def test_um_normalise_callback_missing_entry(cube, caplog):
    """Warning when STASH dictionary doesn't contain stash."""
    cube.attributes["STASH"] = "m00s00i000"
    read._um_normalise_callback(cube, None, None)
    _, level, message = caplog.record_tuples[0]
    assert level == logging.WARNING


def test_lfric_normalise_callback_remove_attrs(cube):
    """Correctly remove unneeded attributes."""
    cube.attributes["uuid"] = "87096862-89c3-4749-9c6c-0be91c2a7954"
    cube.attributes["timeStamp"] = "2024-May-20 12:29:21 GMT"
    read._lfric_normalise_callback(cube, None, None)
    assert "uuid" not in cube.attributes
    assert "timeStamp" not in cube.attributes


def test_lfric_normalise_callback_sort_stash(cube):
    """Correctly sort STASH code lists."""
    cube.attributes["um_stash_source"] = "['m01s03i025', 'm01s00i025']"
    read._lfric_normalise_callback(cube, None, None)
    actual = cube.attributes["um_stash_source"]
    expected = "['m01s00i025', 'm01s03i025']"
    assert actual == expected


def test_lfric_time_coord_fix_callback():
    """Correctly convert time from AuxCoord to DimCoord."""
    time_coord = iris.coords.AuxCoord([0, 1, 2], standard_name="time")
    cube = iris.cube.Cube([0, 0, 0], aux_coords_and_dims=[(time_coord, 0)])
    read._lfric_time_coord_fix_callback(cube, None, None)
    assert isinstance(cube.coord("time"), iris.coords.DimCoord)
    assert cube.coord_dims("time") == (0,)


def test_lfric_time_coord_fix_callback_scalar_time():
    """Correctly convert time from AuxCoord to DimCoord for scalar time."""
    length_coord = iris.coords.DimCoord([0, 1, 2], var_name="length")
    time_coord = iris.coords.AuxCoord([0], standard_name="time")
    cube = iris.cube.Cube([0, 0, 0], aux_coords_and_dims=[(length_coord, 0)])
    cube.add_aux_coord(time_coord)
    read._lfric_time_coord_fix_callback(cube, None, None)
    assert isinstance(cube.coord("time"), iris.coords.AuxCoord)
    assert cube.coord_dims("time") == ()


def test_lfric_time_coord_fix_callback_no_time():
    """Don't do anything if no time coordinate present."""
    length_coord = iris.coords.DimCoord([0, 1, 2], var_name="length")
    cube = iris.cube.Cube([0, 0, 0], aux_coords_and_dims=[(length_coord, 0)])
    read._lfric_time_coord_fix_callback(cube, None, None)
    assert len(cube.coords("time")) == 0


def test_pressure_coord_name_fix_callback(transect_source_cube):
    """Check that pressure_level is renamed to pressure if it exists."""
    cube = transect_source_cube.copy()
    cube.coord("pressure").rename("pressure_level")
    read._fix_pressure_coord_callback(cube)
    assert "pressure" in [coord.name() for coord in cube.coords()]


def test_pressure_coord_unit_fix_callback(transect_source_cube):
    """Check that pressure level units are hPa."""
    cube = transect_source_cube.copy()
    cube.coord("pressure").convert_units("Pa")
    read._fix_pressure_coord_callback(cube)
    assert str(cube.coord("pressure").units) == "hPa"
    assert cube.coord("pressure").points[0] == 100


def test_spatial_coord_rename_callback():
    """Check that spatial coord gets renamed if it is not grid_latitude."""
    # This cube contains 'latitude' and 'longitude'
    cube = iris.load_cube("tests/test_data/transect_test_umpl.nc")
    read._fix_spatial_coord_name_callback(cube)
    assert (
        repr(cube.coords())
        == "[<DimCoord: time / (hours since 1970-01-01 00:00:00)  [...]  shape(2,)>, <DimCoord: pressure / (hPa)  [ 100., 150., ..., 950., 1000.]  shape(16,)>, <DimCoord: grid_latitude / (degrees)  [-10.98, -10.94, ..., -10.82, -10.78]  shape(6,)>, <DimCoord: grid_longitude / (degrees)  [19.02, 19.06, ..., 19.18, 19.22]  shape(6,)>, <DimCoord: forecast_reference_time / (hours since 1970-01-01 00:00:00)  [...]>, <AuxCoord: forecast_period / (hours)  [15., 18.]  shape(2,)>]"
    )


def test_spatial_coord_not_exist_callback():
    """Check that spatial coord returns cube if cube does not contain spatial coordinates."""
    cube = iris.load_cube("tests/test_data/transect_test_umpl.nc")
    cube = cube[:, :, 0, 0]  # Remove spatial dimcoords
    read._fix_spatial_coord_name_callback(cube)
    assert (
        repr(cube.coords())
        == "[<DimCoord: time / (hours since 1970-01-01 00:00:00)  [...]  shape(2,)>, <DimCoord: pressure / (hPa)  [ 100., 150., ..., 950., 1000.]  shape(16,)>, <DimCoord: forecast_reference_time / (hours since 1970-01-01 00:00:00)  [...]>, <DimCoord: latitude / (degrees)  [-10.98]>, <DimCoord: longitude / (degrees)  [19.02]>, <AuxCoord: forecast_period / (hours)  [15., 18.]  shape(2,)>]"
    )


def test_lfric_timecoord_fix_forecastperiod(slammed_lfric_cube):
    """Check that read callback creates an appropriate forecast_period dimension."""
    read._lfric_time_callback(slammed_lfric_cube)
    assert (
        repr(slammed_lfric_cube.coord("forecast_period"))
        == "<AuxCoord: forecast_period / (seconds)  [3600., ...]  shape(6,)>"
    )


def test_lfric_timecoord_fix_forecastreftime(slammed_lfric_cube):
    """Check that read callback creates an appropriate forecast_reference_time coord."""
    read._lfric_time_callback(slammed_lfric_cube)
    assert (
        repr(slammed_lfric_cube.coord("forecast_reference_time"))
        == "<DimCoord: forecast_reference_time / (seconds since 2022-01-01 00:00:00)  [...]>"
    )



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for convection diagnostics."""

import iris
import iris.cube
import numpy as np
import pytest

from CSET.operators import convection, read


@pytest.fixture
def SBCAPE() -> iris.cube.Cube:
    """Get a SBCAPE cube."""
    return read.read_cube("tests/test_data/convection/SBCAPE.nc")


@pytest.fixture
def MUCAPE() -> iris.cube.Cube:
    """Get an MUCAPE cube."""
    return read.read_cube("tests/test_data/convection/MUCAPE.nc")


@pytest.fixture
def MUCIN() -> iris.cube.Cube:
    """Get a MUCIN cube."""
    return read.read_cube("tests/test_data/convection/MUCIN.nc")


@pytest.fixture
def EIB() -> iris.cube.Cube:
    """Get an EIB cube."""
    return read.read_cube("tests/test_data/convection/EIB.nc")


@pytest.fixture
def BLheight() -> iris.cube.Cube:
    """Get a BLheight cube."""
    return read.read_cube("tests/test_data/convection/BLheight.nc")


@pytest.fixture
def orography_2D() -> iris.cube.Cube:
    """Get a 2D Orography cube."""
    return read.read_cube("tests/test_data/convection/Orography2D.nc")


@pytest.fixture
def orography_3D() -> iris.cube.Cube:
    """Get a 3D Orography cube (time)."""
    return read.read_cube("tests/test_data/convection/Orography3D.nc")


@pytest.fixture
def orography_3D_ens() -> iris.cube.Cube:
    """Get a 3D Orography cube (realization)."""
    return read.read_cube("tests/test_data/convection/Orography3D_ens.nc")


@pytest.fixture
def orography_4D() -> iris.cube.Cube:
    """Get a 4D Orography cube (time and realization)."""
    return read.read_cube("tests/test_data/convection/Orography4D.nc")


@pytest.fixture(scope="session")
def cape_ratio_kgo() -> iris.cube.Cube:
    """Known good output for CAPE ratio."""
    return read.read_cube("tests/test_data/convection/ECFlagB.nc")


@pytest.fixture(scope="session")
def inflow_kgo() -> iris.cube.Cube:
    """Known good output for inflow layer properties."""
    return read.read_cube("tests/test_data/convection/ECFlagD.nc")


def test_cape_ratio(SBCAPE, MUCAPE, MUCIN, cape_ratio_kgo):
    """Compare with precalculated ratio KGOs."""
    # Calculate the diagnostic.
    cape_75 = convection.cape_ratio(SBCAPE, MUCAPE, MUCIN)
    # Compare with KGO.
    assert np.allclose(cape_75.data, cape_ratio_kgo.data, atol=1e-5, equal_nan=True)

    # Calculate the diagnostic.
    cape_1p5 = convection.cape_ratio(SBCAPE, MUCAPE, MUCIN, MUCIN_thresh=-1.5)
    # Compare with KGO.
    precalculated_1p5 = read.read_cube("tests/test_data/convection/ECFlagB_2.nc")
    assert np.allclose(cape_1p5.data, precalculated_1p5.data, atol=1e-5, equal_nan=True)


def test_cape_ratio_non_masked_arrays(SBCAPE, MUCAPE, MUCIN, cape_ratio_kgo):
    """Calculate with non-masked arrays and compare with precalculated ratio."""
    # Replace masked values with NaNs.
    SBCAPE.data = SBCAPE.data.filled(np.nan)
    MUCAPE.data = MUCAPE.data.filled(np.nan)
    MUCIN.data = MUCIN.data.filled(np.nan)

    # Calculate the diagnostic.
    cape_75 = convection.cape_ratio(SBCAPE, MUCAPE, MUCIN)

    # Compare with KGO.
    assert np.allclose(cape_75.data, cape_ratio_kgo.data, atol=1e-5, equal_nan=True)


def test_inflow_layer_properties(EIB, BLheight, orography_2D, inflow_kgo):
    """Compare with precalculated properties."""
    inflow_layer_properties = convection.inflow_layer_properties(
        EIB, BLheight, orography_2D
    )
    assert np.allclose(inflow_layer_properties.data, inflow_kgo.data)


def test_inflow_layer_properties_non_masked_arrays(
    EIB, BLheight, orography_2D, inflow_kgo
):
    """Use non-masked data and compare with precalculated properties."""
    # Unmask the data.
    EIB.data = EIB.data.filled(np.nan)
    inflow_layer_properties = convection.inflow_layer_properties(
        EIB, BLheight, orography_2D
    )
    assert np.allclose(inflow_layer_properties.data, inflow_kgo.data)


def test_inflow_layer_properties_3D_orography_time(
    EIB, BLheight, orography_3D, inflow_kgo
):
    """Reduce a 3D orography (time) field down to 2D."""
    inflow_layer_properties = convection.inflow_layer_properties(
        EIB, BLheight, orography_3D
    )
    assert np.allclose(inflow_layer_properties.data, inflow_kgo.data)


def test_inflow_layer_properties_3D_orography_ensemble(
    EIB, BLheight, orography_3D_ens, inflow_kgo
):
    """Reduce a 3D orography (realisation) field down to 2D."""
    inflow_layer_properties = convection.inflow_layer_properties(
        EIB, BLheight, orography_3D_ens
    )
    assert np.allclose(inflow_layer_properties.data, inflow_kgo.data)


def test_inflow_layer_properties_4D_orography(EIB, BLheight, orography_4D, inflow_kgo):
    """Reduce a 4D orography (time and realisation) field down to 2D."""
    inflow_layer_properties = convection.inflow_layer_properties(
        EIB, BLheight, orography_4D
    )
    assert np.allclose(inflow_layer_properties.data, inflow_kgo.data)



#! /usr/bin/env python3

"""Retrieve the files from the filesystem for the current cycle point."""

import abc
import glob
import itertools
import logging
import os
import shutil
import ssl
import sys
import urllib.parse
import urllib.request
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timedelta
from typing import Literal

import isodate

logging.basicConfig(
    level=os.getenv("LOGLEVEL", "INFO"), format="%(asctime)s %(levelname)s %(message)s"
)


class FileRetrieverABC(abc.ABC):
    """Abstract base class for retrieving files from a data source.

    The `get_file` method must be defined. Optionally the __enter__ and __exit__
    methods maybe be overridden to add setup or cleanup code.

    The class is designed to be used as a context manager, so that resources can
    be cleaned up after the retrieval is complete. All the files of a model are
    retrieved within a single context manager block, within which the `get_file`
    method is called for each file path.
    """

    def __enter__(self) -> "FileRetrieverABC":
        """Initialise the file retriever."""
        logging.debug("Initialising FileRetriever.")
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Clean up the file retriever."""
        logging.debug("Tearing down FileRetriever.")

    @abc.abstractmethod
    def get_file(self, file_path: str, output_dir: str) -> bool:  # pragma: no cover
        """Save a file from the data source to the output directory.

        Not all of the given paths will exist, so FileNotFoundErrors should be
        logged, but not raised.

        Implementations should be thread safe, as the method is called from
        multiple threads.

        Parameters
        ----------
        file_path: str
            Path of the file to copy on the data source. It may contain patterns
            like globs, which will be expanded in a system specific manner.
        output_dir: str
            Path to filesystem directory into which the file should be copied.

        Returns
        -------
        bool:
            True if files were transferred, otherwise False.
        """
        raise NotImplementedError


class FilesystemFileRetriever(FileRetrieverABC):
    """Retrieve files from the filesystem."""

    def get_file(self, file_path: str, output_dir: str) -> bool:
        """Save a file from the filesystem to the output directory.

        Parameters
        ----------
        file_path: str
            Path of the file to copy on the filesystem. It may contain patterns
            like globs, which will be expanded in a system specific manner.
        output_dir: str
            Path to filesystem directory into which the file should be copied.

        Returns
        -------
        bool:
            True if files were transferred, otherwise False.
        """
        file_paths = glob.glob(os.path.expanduser(file_path))
        logging.debug("Copying files:\n%s", "\n".join(file_paths))
        if not file_paths:
            logging.warning("file_path does not match any files: %s", file_path)
        any_files_copied = False
        for file in file_paths:
            try:
                shutil.copy(file, output_dir)
                any_files_copied = True
            except OSError as err:
                logging.warning("Failed to copy %s, error: %s", file, err)
        return any_files_copied


class HTTPFileRetriever(FileRetrieverABC):
    """Retrieve files via HTTP."""

    def get_file(self, file_path: str, output_dir: str) -> bool:
        """Save a file from a HTTP address to the output directory.

        Parameters
        ----------
        file_path: str
            Path of the file to copy on MASS. It may contain patterns like
            globs, which will be expanded in a system specific manner.
        output_dir: str
            Path to filesystem directory into which the file should be copied.

        Returns
        -------
        bool:
            True if files were transferred, otherwise False.
        """
        ctx = ssl.create_default_context()
        # Needed to enable compatibility with malformed iBoss TLS certificates.
        ctx.verify_flags &= ~ssl.VERIFY_X509_STRICT
        save_path = (
            f"{output_dir.removesuffix('/')}/"
            + urllib.parse.urlparse(file_path).path.split("/")[-1]
        )
        any_files_copied = False
        try:
            with urllib.request.urlopen(file_path, timeout=30, context=ctx) as response:
                with open(save_path, "wb") as fp:
                    # Read in 1 MiB chunks so data needn't fit in memory.
                    while data := response.read(1024 * 1024):
                        fp.write(data)
                any_files_copied = True
        except OSError as err:
            logging.warning("Failed to retrieve %s, error: %s", file_path, err)
        return any_files_copied


def _get_needed_environment_variables() -> dict:
    """Load the needed variables from the environment."""
    # Python 3.10 and older don't fully support ISO 8601 datetime formats.
    # TODO: Remove once we drop python 3.10.
    if sys.version_info.minor < 11:
        _fromisoformat = isodate.parse_datetime
    else:
        _fromisoformat = datetime.fromisoformat
    variables = {
        "raw_path": os.environ["DATA_PATH"],
        "date_type": os.environ["DATE_TYPE"],
        "data_time": _fromisoformat(os.environ["CYLC_TASK_CYCLE_POINT"]),
        "forecast_length": isodate.parse_duration(os.environ["ANALYSIS_LENGTH"]),
        "forecast_offset": isodate.parse_duration(os.environ["ANALYSIS_OFFSET"]),
        "model_identifier": os.environ["MODEL_IDENTIFIER"],
        "rose_datac": os.environ["ROSE_DATAC"],
    }
    try:
        variables["data_period"] = isodate.parse_duration(os.environ["DATA_PERIOD"])
    except KeyError:
        # Data period is not needed for initiation time.
        if variables["date_type"] != "initiation":
            raise
        variables["data_period"] = None
    logging.debug("Environment variables loaded: %s", variables)
    return variables


def _template_file_path(
    raw_path: str,
    date_type: Literal["validity", "initiation"],
    data_time: datetime,
    forecast_length: timedelta,
    forecast_offset: timedelta,
    data_period: timedelta,
) -> list[str]:
    """Fill time placeholders to generate a file path to fetch."""
    placeholder_times: list[datetime] = []
    lead_times: list[timedelta] = []
    match date_type:
        case "validity":
            date = data_time
            while date < data_time + forecast_length:
                placeholder_times.append(date)
                date += data_period
        case "initiation":
            placeholder_times.append(data_time)
            lead_time = forecast_offset
            while lead_time < forecast_length:
                lead_times.append(lead_time)
                lead_time += data_period
        case _:
            raise ValueError(f"Invalid date type: {date_type}")

    paths: set[str] = set()
    for placeholder_time in placeholder_times:
        # Expand out all other format strings.
        path = placeholder_time.strftime(os.path.expandvars(raw_path))
        if lead_times:
            # Expand out lead time format strings, %N.
            for lead_time in lead_times:
                # BUG: Will not respect escaped % signs, e.g: %%N.
                paths.add(
                    path.replace("%N", f"{int(lead_time.total_seconds()) // 3600:03d}")
                )
        else:
            paths.add(path)
    return sorted(paths)


def fetch_data(file_retriever: FileRetrieverABC = FilesystemFileRetriever):
    """Fetch the data for a model.

    The following environment variables need to be set:
    * ANALYSIS_OFFSET
    * ANALYSIS_LENGTH
    * CYLC_TASK_CYCLE_POINT
    * DATA_PATH
    * DATA_PERIOD
    * DATE_TYPE
    * MODEL_IDENTIFIER
    * ROSE_DATAC

    Parameters
    ----------
    file_retriever: FileRetriever
        FileRetriever implementation to use. Defaults to FilesystemFileRetriever.

    Raises
    ------
    FileNotFound:
        If no files are found for the model, across all tried paths.
    """
    v = _get_needed_environment_variables()

    # Prepare output directory.
    cycle_data_dir = f"{v['rose_datac']}/data/{v['model_identifier']}"
    os.makedirs(cycle_data_dir, exist_ok=True)
    logging.debug("Output directory: %s", cycle_data_dir)

    # Get file paths.
    paths = _template_file_path(
        v["raw_path"],
        v["date_type"],
        v["data_time"],
        v["forecast_length"],
        v["forecast_offset"],
        v["data_period"],
    )
    logging.info("Retrieving paths:\n%s", "\n".join(paths))

    # Use file retriever to transfer data with multiple threads.
    with file_retriever() as retriever, ThreadPoolExecutor() as executor:
        files_found = any(
            executor.map(retriever.get_file, paths, itertools.repeat(cycle_data_dir))
        )
    # We don't need to exhaust the iterator, as all futures are submitted
    # before map yields anything. Therefore they will all be resolved upon
    # exiting the with block.
    if not files_found:
        raise FileNotFoundError("No files found for model!")



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests transect operator."""

from pathlib import Path

import iris
import iris.coord_systems
import iris.cube
import numpy as np
import pytest

from CSET.operators import plot, read, transect


# Session scope fixtures, so the test data only has to be loaded once.
@pytest.fixture(scope="session")
def transect_source_cube_out() -> iris.cube.Cube:
    """Load cube containing UM pressure level data transect."""
    return read.read_cube("tests/test_data/transect_out_umpl.nc")


@pytest.fixture(scope="session")
def load_cube_ml() -> iris.cube.Cube:
    """Load cube containing UM model level data."""
    return read.read_cube("tests/test_data/transect_test_umml.nc")


@pytest.fixture(scope="session")
def load_cube_ml_out() -> iris.cube.Cube:
    """Load cube containing UM model level data transect."""
    return read.read_cube("tests/test_data/transect_out_umml.nc")


@pytest.fixture(scope="session")
def test_transect_pl(transect_source_cube, transect_source_cube_out):
    """Test case of computing transect on UM pressure level data."""
    assert np.allclose(
        transect.calc_transect(
            transect_source_cube, startcoords=(-10.94, 19.06), endcoords=(-10.82, 19.18)
        ).data,
        transect_source_cube_out.data,
        rtol=1e-06,
        atol=1e-02,
    )


@pytest.fixture(scope="session")
def test_transect_ml(load_cube_ml, load_cube_ml_out):
    """Test case of computing transect on UM model level data."""
    assert np.allclose(
        transect.calc_transect(
            load_cube_ml, startcoords=(-0.94, 29.06), endcoords=(-0.78, 29.3)
        ).data,
        load_cube_ml_out.data,
        rtol=1e-06,
        atol=1e-02,
    )


def test_transect_45deg(transect_source_cube, transect_source_cube_out):
    """Test case of 45 degree angle where map coordinate should be longitude."""
    with pytest.raises(iris.exceptions.CoordinateNotFoundError):
        transect.calc_transect(
            transect_source_cube, startcoords=(-10.94, 19.06), endcoords=(-10.86, 19.14)
        ).coord("latitude")


def test_transect_coord_outofboundsLLat(transect_source_cube):
    """Test case of computing transect on coords out of range (low start lat)."""
    with pytest.raises(IndexError):
        transect.calc_transect(
            transect_source_cube, startcoords=(-11.94, 19.06), endcoords=(-10.82, 19.18)
        )


def test_transect_coord_outofboundsLLon(transect_source_cube):
    """Test case of computing transect on coords out of range (low start lon)."""
    with pytest.raises(IndexError):
        transect.calc_transect(
            transect_source_cube, startcoords=(-10.94, 10.06), endcoords=(-10.82, 19.18)
        )


def test_transect_coord_outofboundsHLat(transect_source_cube):
    """Test case of computing transect on coords out of range (high end lat)."""
    with pytest.raises(IndexError):
        transect.calc_transect(
            transect_source_cube, startcoords=(-10.94, 19.06), endcoords=(-5.82, 19.18)
        )


def test_transect_coord_outofboundsHLon(transect_source_cube):
    """Test case of computing transect on coords out of range (high end lon)."""
    with pytest.raises(IndexError):
        transect.calc_transect(
            transect_source_cube, startcoords=(-10.94, 19.06), endcoords=(-10.82, 25.18)
        )


def test_transect_90degangle(transect_source_cube):
    """Test case of computing transect on 90 degree angle (no delta lon)."""
    with pytest.raises(iris.exceptions.CoordinateNotFoundError):
        transect.calc_transect(
            transect_source_cube, startcoords=(-10.94, 19.18), endcoords=(-10.82, 19.18)
        ).coord("longitude")


def test_transect_180degangle(transect_source_cube):
    """Test case of computing transect on 180 degree angle (no delta lat)."""
    with pytest.raises(iris.exceptions.CoordinateNotFoundError):
        transect.calc_transect(
            transect_source_cube, startcoords=(-10.94, 19.06), endcoords=(-10.94, 19.18)
        ).coord("latitude")


def test_transect_plotasfuncoflatitude(transect_source_cube):
    """Test case of computing transect where it should return as function of latitude."""
    with pytest.raises(iris.exceptions.CoordinateNotFoundError):
        transect.calc_transect(
            transect_source_cube, startcoords=(-10.94, 19.06), endcoords=(-10.82, 19.14)
        ).coord("longitude")


def test_transect_plotasfuncoflongitude(transect_source_cube):
    """Test case of computing transect where it should return as function of longitude."""
    with pytest.raises(iris.exceptions.CoordinateNotFoundError):
        transect.calc_transect(
            transect_source_cube, startcoords=(-10.94, 19.06), endcoords=(-10.86, 19.18)
        ).coord("latitude")


def test_transect_model_level_spatial_contour_plot(load_cube_ml_out, tmp_working_dir):
    """Plot a contour plot of the transect model level data."""
    plot.spatial_contour_plot(load_cube_ml_out, filename="plot")
    assert Path("plot_449743.0.png").is_file()
    assert Path("plot_449744.0.png").is_file()


def test_transect_model_level_spatial_pcolormesh_plot(
    load_cube_ml_out, tmp_working_dir
):
    """Plot a pcolormesh plot of the transect model level data."""
    plot.spatial_pcolormesh_plot(load_cube_ml_out, filename="plot")
    assert Path("plot_449743.0.png").is_file()
    assert Path("plot_449744.0.png").is_file()


def test_transect_pressure_spatial_contour_plot(
    transect_source_cube_out, tmp_working_dir
):
    """Plot a contour plot of the transect pressure data."""
    plot.spatial_contour_plot(transect_source_cube_out, filename="plot")
    assert Path("plot_449469.0.png").is_file()
    assert Path("plot_449472.0.png").is_file()


def test_transect_pressure_spatial_pcolormesh_plot(
    transect_source_cube_out, tmp_working_dir
):
    """Plot a pcolormesh plot of the transect pressure data."""
    plot.spatial_pcolormesh_plot(transect_source_cube_out, filename="plot")
    assert Path("plot_449469.0.png").is_file()
    assert Path("plot_449472.0.png").is_file()



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CSET: Convective and turbulence Scale Evaluation Tool."""

import argparse
import logging
import os
import shlex
import sys
from importlib.metadata import version
from pathlib import Path

from CSET._common import ArgumentError


def main(raw_cli_args: list[str] = sys.argv):
    """CLI entrypoint.

    Handles argument parsing, setting up logging, top level error capturing,
    and execution of the desired subcommand.
    """
    # Read arguments from the command line and CSET_ADDOPTS environment variable
    # into an args object.
    parser = setup_argument_parser()
    cli_args = raw_cli_args[1:] + shlex.split(os.getenv("CSET_ADDOPTS", ""))
    args, unparsed_args = parser.parse_known_args(cli_args)

    setup_logging(args.verbose)

    # Down here so runs after logging is setup.
    logging.debug("CLI Arguments: %s", cli_args)

    if args.subparser is None:
        print("Please choose a command.", file=sys.stderr)
        parser.print_usage()
        sys.exit(127)

    try:
        # Execute the specified subcommand.
        args.func(args, unparsed_args)
    except ArgumentError as err:
        # Error message for when needed template variables are missing.
        print(err, file=sys.stderr)
        parser.print_usage()
        sys.exit(127)
    except Exception as err:
        # Provide slightly nicer error messages for unhandled exceptions.
        print(err, file=sys.stderr)
        # Display the time and full traceback when debug logging.
        logging.debug("An unhandled exception occurred.")
        if logging.root.isEnabledFor(logging.DEBUG):
            raise
        sys.exit(1)


def setup_argument_parser() -> argparse.ArgumentParser:
    """Create argument parser for CSET CLI."""
    parser = argparse.ArgumentParser(
        prog="cset", description="Convective Scale Evaluation Tool"
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="increase output verbosity, may be specified multiple times",
    )
    parser.add_argument(
        "--version", action="version", version=f"CSET v{version('CSET')}"
    )

    # https://docs.python.org/3/library/argparse.html#sub-commands
    subparsers = parser.add_subparsers(title="subcommands", dest="subparser")

    # Run operator chain
    parser_bake = subparsers.add_parser("bake", help="run steps from a recipe file")
    parser_bake.add_argument(
        "-i",
        "--input-dir",
        type=str,
        action="extend",
        required=True,
        nargs="+",
        help="directories containing input data. May contain wildcards",
    )
    parser_bake.add_argument(
        "-o",
        "--output-dir",
        type=Path,
        required=True,
        help="directory to write output into",
    )
    parser_bake.add_argument(
        "-r",
        "--recipe",
        type=Path,
        required=True,
        help="recipe file to read",
    )
    parser_bake.add_argument(
        "-s", "--style-file", type=Path, help="colour bar definition to use"
    )
    parser_bake.add_argument(
        "--plot-resolution", type=int, help="plotting resolution in dpi"
    )
    parser_bake.set_defaults(func=_bake_command)

    parser_graph = subparsers.add_parser("graph", help="visualise a recipe file")
    parser_graph.add_argument(
        "-d",
        "--details",
        action="store_true",
        help="include operator arguments in output",
    )
    parser_graph.add_argument(
        "-o",
        "--output-path",
        type=Path,
        nargs="?",
        help="persistent file to save the graph. Otherwise the file is opened",
        default=None,
    )
    parser_graph.add_argument(
        "-r",
        "--recipe",
        type=Path,
        required=True,
        help="recipe file to read",
    )
    parser_graph.set_defaults(func=_graph_command)

    parser_cookbook = subparsers.add_parser(
        "cookbook", help="unpack included recipes to a folder"
    )
    parser_cookbook.add_argument(
        "-d",
        "--details",
        action="store_true",
        help="list available recipes. Supplied recipes are detailed",
    )
    parser_cookbook.add_argument(
        "-o",
        "--output-dir",
        type=Path,
        help="directory to save recipes. If omitted uses $PWD",
        default=Path.cwd(),
    )
    parser_cookbook.add_argument(
        "recipe",
        type=str,
        nargs="?",
        help="recipe to output or detail",
        default="",
    )
    parser_cookbook.set_defaults(func=_cookbook_command)

    parser_recipe_id = subparsers.add_parser("recipe-id", help="get the ID of a recipe")
    parser_recipe_id.add_argument(
        "-r",
        "--recipe",
        type=Path,
        required=True,
        help="recipe file to read",
    )
    parser_recipe_id.set_defaults(func=_recipe_id_command)

    return parser


def setup_logging(verbosity: int):
    """Configure logging level, format and output stream.

    Level is based on verbose argument and the LOGLEVEL environment variable.
    """
    logging.captureWarnings(True)

    # Calculate logging level.
    try:
        name_to_level = logging.getLevelNamesMapping()
    except AttributeError:
        # logging.getLevelNamesMapping() is python 3.11 or newer. Using
        # implementation detail for older versions.
        name_to_level = logging._nameToLevel
    # Level from CLI flags.
    if verbosity >= 2:
        cli_loglevel = logging.DEBUG
    elif verbosity == 1:
        cli_loglevel = logging.INFO
    else:
        cli_loglevel = logging.WARNING
    # Level from environment variable.
    env_loglevel = name_to_level.get(os.getenv("LOGLEVEL"), logging.ERROR)
    loglevel = min(cli_loglevel, env_loglevel)

    # Configure the root logger.
    logger = logging.getLogger()
    # Record everything at least INFO for the log file.
    logger.setLevel(min(loglevel, logging.INFO))
    # Hide matplotlib's many font messages.
    logger.addFilter(
        lambda record: not (
            isinstance(record, str) and record.msg.startswith("findfont:")
        )
    )
    stderr_log = logging.StreamHandler()
    # Filter stderr log to just what is requested.
    stderr_log.addFilter(lambda record: record.levelno >= loglevel)
    stderr_log.addFilter(
        lambda record: not (
            isinstance(record, str) and record.msg.startswith("findfont:")
        )
    )
    stderr_log.setFormatter(logging.Formatter("%(asctime)s %(levelname)s %(message)s"))
    logger.addHandler(stderr_log)


def _bake_command(args, unparsed_args):
    from CSET._common import parse_variable_options
    from CSET.operators import execute_recipe

    recipe_variables = parse_variable_options(unparsed_args)
    execute_recipe(
        args.recipe,
        args.input_dir,
        args.output_dir,
        recipe_variables,
        args.style_file,
        args.plot_resolution,
    )


def _graph_command(args, unparsed_args):
    from CSET.graph import save_graph

    save_graph(
        args.recipe,
        args.output_path,
        auto_open=not args.output_path,
        detailed=args.details,
    )


def _cookbook_command(args, unparsed_args):
    from CSET.recipes import detail_recipe, list_available_recipes, unpack_recipe

    if args.recipe:
        if args.details:
            detail_recipe(args.recipe)
        else:
            try:
                unpack_recipe(args.output_dir, args.recipe)
            except FileNotFoundError:
                logging.error("Recipe %s does not exist.", args.recipe)
                sys.exit(1)
    else:
        list_available_recipes()


def _recipe_id_command(args, unparsed_args):
    from uuid import uuid4

    from CSET._common import parse_recipe, parse_variable_options, slugify

    recipe_variables = parse_variable_options(unparsed_args)
    recipe = parse_recipe(args.recipe, recipe_variables)
    try:
        recipe_id = slugify(recipe["title"])
    except KeyError:
        logging.warning("Recipe has no title; Falling back to random recipe_id.")
        recipe_id = str(uuid4())
    print(recipe_id)



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Operators to extract a transect given a tuple of xy coords to start/finish."""

import logging

import iris
import numpy as np

from CSET.operators._utils import get_cube_yxcoordname


def _check_within_bounds(point: tuple[float, float], lat_coord, lon_coord):
    """Check if the point (lat, lon) is within the bounds of the data."""
    lon_min = min(lon_coord.points)
    lon_max = max(lon_coord.points)
    lat_min = min(lat_coord.points)
    lat_max = max(lat_coord.points)
    if (
        lat_min > point[0]
        or lat_max < point[0]
        or lon_min > point[1]
        or lon_max < point[1]
    ):
        raise IndexError(
            f"Point {point} not between {(lat_min, lon_min)} and {(lat_max, lon_max)}"
        )


def calc_transect(cube: iris.cube.Cube, startcoords: tuple, endcoords: tuple):
    """Compute transect between startcoords and endcoords.

    Computes a transect for a given cube containing at least latitude
    and longitude coordinates, using an appropriate sampling interval along the
    transect based on grid spacing. Also decides a suitable x coordinate to plot along
    the transect - see notes for details on this.

    Arguments
    ---------

    cube: Cube
        An iris cube containing latitude and longitude coordinate dimensions,
        to compute the transect on.
    startcoords: tuple
        A tuple containing the start coordinates for the transect using model coordinates,
        ordered (latitude,longitude).
    endcoords: tuple
        A tuple containing the end coordinates for the transect using model coordinates,
        ordered (latitude,longitude).

    Returns
    -------
    cube: Cube
        A cube containing at least pressure and the coordinate specified by coord, for
        the transect specified between startcoords and endcoords.

    Notes
    -----
    This operator uses the iris.Nearest method to interpolate the specific point along
    the transect.
    Identification of an appropriate coordinate to plot along the x axis is done by
    determining the largest distance out of latitude and longitude. For example, if
    the transect is 90 degrees (west to east), then delta latitude is zero, so it will
    always plot as a function of longitude. Vice versa if the transect is 180 degrees
    (south to north). If a transect is 45 degrees, and delta latitude and longitude
    are the same, it will default to plotting as a function of longitude. Note this
    doesn't affect the transect plot, its purely for interpretation with some appropriate
    x axis labelling/points.
    """
    # Find out xy coord name
    lat_name, lon_name = get_cube_yxcoordname(cube)

    lon_coord = cube.coord(lon_name)
    lat_coord = cube.coord(lat_name)

    _check_within_bounds(startcoords, lat_coord, lon_coord)
    _check_within_bounds(endcoords, lat_coord, lon_coord)

    # Compute vector distance between start and end points in degrees.
    dist_deg = np.sqrt(
        (startcoords[0] - endcoords[0]) ** 2 + (startcoords[1] - endcoords[1]) ** 2
    )

    # Compute minimum gap between x/y spatial coords.
    lon_min = np.abs(np.min(lon_coord.points[1:] - lon_coord.points[:-1]))
    lat_min = np.abs(np.min(lat_coord.points[1:] - lat_coord.points[:-1]))

    # For scenarios where coord is at 90 degree to the grid (i.e. no
    # latitude/longitude change). Only xmin or ymin will be zero, not both
    # (otherwise startcoords and endcoords the same).
    if startcoords[1] == endcoords[1]:
        # Along latitude.
        lon_pnts = np.repeat(startcoords[1], int(dist_deg / lat_min))
        lat_pnts = np.linspace(startcoords[0], endcoords[0], int(dist_deg / lat_min))
        transect_coord = "latitude"
    elif startcoords[0] == endcoords[0]:
        # Along longitude.
        lon_pnts = np.linspace(startcoords[1], endcoords[1], int(dist_deg / lon_min))
        lat_pnts = np.repeat(startcoords[0], int(dist_deg / lon_min))
        transect_coord = "longitude"
    else:
        # Else use the smallest grid space in x or y direction.
        number_of_points = int(dist_deg / np.min([lon_min, lat_min]))
        lon_pnts = np.linspace(startcoords[1], endcoords[1], number_of_points)
        lat_pnts = np.linspace(startcoords[0], endcoords[0], number_of_points)

        # If change in latitude larger than change in longitude:
        if abs(startcoords[0] - endcoords[0]) > abs(startcoords[1] - endcoords[1]):
            transect_coord = "latitude"
        elif abs(startcoords[0] - endcoords[0]) <= abs(startcoords[1] - endcoords[1]):
            transect_coord = "longitude"

    # Create cubelist to store interpolated points along transect.
    interpolated_cubes = iris.cube.CubeList()

    # Iterate over all points along transect, lon_pnts will be the same shape as
    # lat_pnts so we can use either to iterate over.
    for i in range(0, lon_pnts.shape[0]):
        logging.info("%s/%s", i + 1, lon_pnts.shape[0])

        # Get point along transect.
        cube_slice = cube.interpolate(
            [(lon_name, lon_pnts[i]), (lat_name, lat_pnts[i])], iris.analysis.Nearest()
        )

        # Remove existing coordinates ready to add one single map coordinate
        cube_slice.remove_coord(lon_name)
        cube_slice.remove_coord(lat_name)

        if transect_coord == "latitude":
            dist_coord = iris.coords.DimCoord(
                lat_pnts[i], long_name="latitude", units="degrees"
            )
            cube_slice.add_aux_coord(dist_coord)
            cube_slice = iris.util.new_axis(cube_slice, scalar_coord="latitude")
        elif transect_coord == "longitude":
            dist_coord = iris.coords.DimCoord(
                lon_pnts[i], long_name="longitude", units="degrees"
            )
            cube_slice.add_aux_coord(dist_coord)
            cube_slice = iris.util.new_axis(cube_slice, scalar_coord="longitude")

        interpolated_cubes.append(cube_slice)

    # Concatenate into single cube.
    interpolated_cubes = interpolated_cubes.concatenate()

    # Add metadata to interpolated cubes showing coordinates.
    interpolated_cubes[0].attributes["transect_coords"] = (
        f"{startcoords[0]}_{startcoords[1]}_{endcoords[1]}_{endcoords[1]}"
    )
    # If concatenation successful, should be CubeList with one cube left.
    assert len(interpolated_cubes) == 1
    return interpolated_cubes[0]



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Operators to generate constraints to filter with."""

import numbers
import re
from collections.abc import Iterable
from datetime import datetime

import iris
import iris.cube
import iris.exceptions


def generate_stash_constraint(stash: str, **kwargs) -> iris.AttributeConstraint:
    """Generate constraint from STASH code.

    Operator that takes a stash string, and uses iris to generate a constraint
    to be passed into the read operator to minimize the CubeList the read
    operator loads and speed up loading.

    Arguments
    ---------
    stash: str
        stash code to build iris constraint, such as "m01s03i236"

    Returns
    -------
    stash_constraint: iris.AttributeConstraint
    """
    # At a later stage str list an option to combine constraints. Arguments
    # could be a list of stash codes that combined build the constraint.
    stash_constraint = iris.AttributeConstraint(STASH=stash)
    return stash_constraint


def generate_var_constraint(varname: str, **kwargs) -> iris.Constraint:
    """Generate constraint from variable name or STASH code.

    Operator that takes a CF compliant variable name string, and generates an
    iris constraint to be passed into the read or filter operator. Can also be
    passed a STASH code to generate a STASH constraint.

    Arguments
    ---------
    varname: str
        CF compliant name of variable, or a UM STASH code such as "m01s03i236".

    Returns
    -------
    varname_constraint: iris.Constraint
    """
    if re.match(r"m[0-9]{2}s[0-9]{2}i[0-9]{3}$", varname):
        varname_constraint = iris.AttributeConstraint(STASH=varname)
    else:
        varname_constraint = iris.Constraint(name=varname)
    return varname_constraint


def generate_level_constraint(
    coordinate: str, levels: int | list[int] | str, **kwargs
) -> iris.Constraint:
    """Generate constraint for particular levels on the specified coordinate.

    Operator that generates a constraint to constrain to specific model or
    pressure levels. If no levels are specified then any cube with the specified
    coordinate is rejected.

    Typically ``coordinate`` will be ``"pressure"`` or ``"model_level_number"``
    for UM, or ``"full_levels"`` or ``"half_levels"`` for LFRic.

    Arguments
    ---------
    coordinate: str
        Level coordinate name about which to constraint.
    levels: int | list[int] | str
        CF compliant level points, ``"*"`` for retrieving all levels, or
        ``[]`` for no levels.

    Returns
    -------
    constraint: iris.Constraint

    Notes
    -----
    Due to the specification of ``coordinate`` as an argument any iterable
    coordinate can be stratified with this function. Therefore,
    ``"realization"`` is a valid option. Subsequently, ``levels`` specifies the
    ensemble members, or group of ensemble members you wish to constrain your
    results over.
    """
    # If asterisks, then return all levels for given coordinate.
    if levels == "*":
        return iris.Constraint(**{coordinate: lambda cell: True})
    else:
        # Ensure is iterable.
        if not isinstance(levels, Iterable):
            levels = [levels]

        # When no levels specified reject cube with level coordinate.
        if len(levels) == 0:

            def no_levels(cube):
                # Reject cubes for which coordinate exists.
                return not cube.coords(coordinate)

            return iris.Constraint(cube_func=no_levels)

        # Filter the coordinate to the desired levels.
        # Dictionary unpacking is used to provide programmatic keyword arguments.
        return iris.Constraint(**{coordinate: levels})


def generate_cell_methods_constraint(cell_methods: list, **kwargs) -> iris.Constraint:
    """Generate constraint from cell methods.

    Operator that takes a list of cell methods and generates a constraint from
    that. Use [] to specify non-aggregated data.

    Arguments
    ---------
    cell_methods: list
        cube.cell_methods for filtering

    Returns
    -------
    cell_method_constraint: iris.Constraint
    """
    if len(cell_methods) == 0:

        def check_no_aggregation(cube: iris.cube.Cube) -> bool:
            """Check that any cell methods are "point", meaning no aggregation."""
            return set(cm.method for cm in cube.cell_methods) <= {"point"}

        cell_methods_constraint = iris.Constraint(cube_func=check_no_aggregation)
    else:

        def check_cell_methods(cube: iris.cube.Cube) -> bool:
            return cube.cell_methods == tuple(cell_methods)

        cell_methods_constraint = iris.Constraint(cube_func=check_cell_methods)
    return cell_methods_constraint


def generate_time_constraint(
    time_start: str, time_end: str = None, **kwargs
) -> iris.AttributeConstraint:
    """Generate constraint between times.

    Operator that takes one or two ISO 8601 date strings, and returns a
    constraint that selects values between those dates (inclusive).

    Arguments
    ---------
    time_start: str | datetime.datetime
        ISO date for lower bound

    time_end: str | datetime.datetime
        ISO date for upper bound. If omitted it defaults to the same as
        time_start

    Returns
    -------
    time_constraint: iris.Constraint
    """
    if isinstance(time_start, str):
        time_start = datetime.fromisoformat(time_start)
    if time_end is None:
        time_end = time_start
    elif isinstance(time_end, str):
        time_end = datetime.fromisoformat(time_end)
    time_constraint = iris.Constraint(time=lambda t: time_start <= t.point <= time_end)
    return time_constraint


def generate_area_constraint(
    lat_start: float | None,
    lat_end: float | None,
    lon_start: float | None,
    lon_end: float | None,
    **kwargs,
) -> iris.Constraint:
    """Generate an area constraint between latitude/longitude limits.

    Operator that takes a set of latitude and longitude limits and returns a
    constraint that selects grid values only inside that area. Works with the
    data's native grid so is defined within the rotated pole CRS.

    Alternatively, all arguments may be None to indicate the area should not be
    constrained. This is useful to allow making subsetting an optional step in a
    processing pipeline.

    Arguments
    ---------
    lat_start: float | None
        Latitude value for lower bound
    lat_end: float | None
        Latitude value for top bound
    lon_start: float | None
        Longitude value for left bound
    lon_end: float | None
        Longitude value for right bound

    Returns
    -------
    area_constraint: iris.Constraint
    """
    # Check all arguments are defined, or all are None.
    if not (
        all(
            (
                isinstance(lat_start, numbers.Real),
                isinstance(lat_end, numbers.Real),
                isinstance(lon_start, numbers.Real),
                isinstance(lon_end, numbers.Real),
            )
        )
        or all((lat_start is None, lat_end is None, lon_start is None, lon_end is None))
    ):
        raise TypeError("Bounds must real numbers, or all None.")

    # Don't constrain area if all arguments are None.
    if lat_start is None:  # Only need to check once, as they will be the same.
        # An empty constraint allows everything.
        return iris.Constraint()

    # Handle bounds crossing the date line.
    if lon_end < lon_start:
        lon_end = lon_end + 360

    def bound_lat(cell: iris.coords.Cell) -> bool:
        return lat_start < cell < lat_end

    def bound_lon(cell: iris.coords.Cell) -> bool:
        # Adjust cell values to handle crossing the date line.
        if cell < lon_start:
            cell = cell + 360
        return lon_start < cell < lon_end

    area_constraint = iris.Constraint(
        coord_values={"grid_latitude": bound_lat, "grid_longitude": bound_lon}
    )
    return area_constraint


def combine_constraints(
    constraint: iris.Constraint = None, **kwargs
) -> iris.Constraint:
    """
    Operator that combines multiple constraints into one.

    Arguments
    ---------
    constraint: iris.Constraint
        First constraint to combine.
    additional_constraint_1: iris.Constraint
        Second constraint to combine. This must be a named argument.
    additional_constraint_2: iris.Constraint
        There can be any number of additional constraint, they just need unique
        names.
    ...

    Returns
    -------
    combined_constraint: iris.Constraint

    Raises
    ------
    TypeError
        If the provided arguments are not constraints.
    """
    # If the first argument is not a constraint, it is ignored. This handles the
    # automatic passing of the previous step's output.
    if isinstance(constraint, iris.Constraint):
        combined_constraint = constraint
    else:
        combined_constraint = iris.Constraint()

    for constr in kwargs.values():
        combined_constraint = combined_constraint & constr
    return combined_constraint



# © Crown copyright, Met Office (2022-2025) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for ensemble operators."""

import iris
import iris.cube
import numpy as np
import pytest

from CSET.operators import ensembles, read


@pytest.fixture()
def xwind() -> iris.cube.Cube:
    """Get xwind to run tests on."""
    return read.read_cube("tests/test_data/ageofair/aoa_in_ens.nc", "x_wind")


@pytest.fixture()
def xwind_deterministic() -> iris.cube.Cube:
    """Get xwind to run tests on."""
    return read.read_cube("tests/test_data/ageofair/aoa_in_rgd.nc", "x_wind")


@pytest.fixture()
def ywind() -> iris.cube.Cube:
    """Get ywind to run tests on."""
    return read.read_cube("tests/test_data/ageofair/aoa_in_ens.nc", "y_wind")


@pytest.fixture()
def ywind_deterministic() -> iris.cube.Cube:
    """Get xwind to run tests on."""
    return read.read_cube("tests/test_data/ageofair/aoa_in_rgd.nc", "y_wind")


def test_DKE(xwind, ywind):
    """Test DKE calculation."""
    DKE_calc = ywind[1:, :].copy()
    DKE_calc.data[:] = 0.0
    DKE_calc.data[:] = (
        0.5 * (xwind.data[0, :] - xwind.data[1:, :]) ** 2
        + 0.5 * (ywind.data[0, :] - ywind.data[1:, :]) ** 2
    )
    assert np.allclose(
        DKE_calc.data,
        ensembles.DKE(xwind, ywind).data,
        rtol=1e-06,
        atol=1e-02,
    )


def test_DKE_no_realization_u(xwind_deterministic, ywind):
    """Test DKE fails with no realization coordinate for u."""
    with pytest.raises(ValueError):
        ensembles.DKE(xwind_deterministic, ywind)


def test_DKE_no_realization_v(xwind, ywind_deterministic):
    """Test DKE fails with no realization coordinate for v."""
    with pytest.raises(ValueError):
        ensembles.DKE(xwind, ywind_deterministic)


def test_DKE_wrong_order_u(xwind, ywind):
    """Test DKE fails when realization is not the first coordinate in u."""
    xwind.transpose([1, 0, 2, 3, 4])
    with pytest.raises(ValueError):
        ensembles.DKE(xwind, ywind)


def test_DKE_wrong_order_v(xwind, ywind):
    """Test DKE fails when realization is not the first coordinate in v."""
    ywind.transpose([1, 0, 2, 3, 4])
    with pytest.raises(ValueError):
        ensembles.DKE(xwind, ywind)


def test_DKE_shape_mismatch(xwind, ywind):
    """Test DKE fails when there is a shape mismatch between u and v."""
    xwind = xwind[:, 1:, :]
    with pytest.raises(ValueError):
        ensembles.DKE(xwind, ywind)


def test_DKE_coordinate_mismatch(xwind, ywind):
    """Test DKE fails when there is a coordinate mismatch between u and v."""
    xwind_new = iris.cube.Cube(
        xwind.data,
        dim_coords_and_dims=[
            (xwind.coord("realization"), 0),
            (xwind.coord("time"), 1),
            (xwind.coord("pressure")[::-1], 2),
            (xwind.coord("grid_latitude"), 3),
            (xwind.coord("grid_longitude"), 4),
        ],
    )
    with pytest.raises(ValueError):
        ensembles.DKE(xwind_new, ywind)



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Operators for writing various types of files to disk."""

import secrets
from pathlib import Path

import iris
import iris.cube

from CSET._common import get_recipe_metadata, slugify


def write_cube_to_nc(
    cube: iris.cube.Cube | iris.cube.CubeList,
    filename: str = None,
    overwrite: bool = False,
    **kwargs,
) -> str:
    """Write a cube to a NetCDF file.

    This operator expects an iris cube object that will then be saved to disk.

    Arguments
    ---------
    cube: iris.cube.Cube | iris.cube.CubeList
        Data to save.
    filename: str, optional
        Path to save the cubes too. Defaults to the recipe title + .nc
    overwrite: bool, optional
        Whether to overwrite an existing file. If False the filename will have a
        unique suffix added. Defaults to False.

    Returns
    -------
    Cube | CubeList
        The inputted cube(list) (so further operations can be applied)
    """
    if filename is None:
        filename = slugify(get_recipe_metadata().get("title", "Untitled"))

    # Append a unique suffix if not overwriting. We use randomness rather than a
    # sequence number to avoid race conditions with multiple job runners.
    if not overwrite:
        filename = f"{filename}_{secrets.token_urlsafe(16)}.nc"

    # Ensure that output filename is a Path with a .nc suffix
    filename = Path(filename).with_suffix(".nc")
    # Save the file as nc compliant (iris should handle this)
    iris.save(cube, filename, zlib=True)
    return cube



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Writing operator tests."""

from pathlib import Path

from CSET.operators import write


def test_write_cube(tmp_path: Path, cube):
    """Write cube and verify it was written."""
    file_path = tmp_path / "cube.nc"
    write.write_cube_to_nc(cube, file_path, overwrite=True)
    assert file_path.is_file()


def test_write_cube_default_filename(cube, tmp_working_dir):
    """Write cube without specifying a filename."""
    Path("meta.json").write_text("{}", encoding="UTF-8")
    write.write_cube_to_nc(cube, overwrite=True)
    assert Path.cwd().joinpath("untitled.nc").is_file()



#! /usr/bin/env python3

"""Run a recipe with the CSET CLI."""

import logging
import os
import subprocess
import sys
import zipfile
from pathlib import Path

logging.basicConfig(
    level=os.getenv("LOGLEVEL", "INFO"), format="%(asctime)s %(levelname)s %(message)s"
)


def subprocess_env():
    """Create a dictionary of amended environment variables for subprocess."""
    env_mapping = dict(os.environ)
    return env_mapping


def recipe_file() -> str:
    """Write the recipe file to disk and return its path as a string."""
    # Ready recipe file to disk.
    cset_recipe = os.environ["CSET_RECIPE_NAME"]
    subprocess.run(("cset", "-v", "cookbook", cset_recipe), check=True)
    # Debug check that recipe has been retrieved.
    assert Path(cset_recipe).is_file()
    return cset_recipe


def recipe_id():
    """Get the ID for the recipe."""
    file = recipe_file()
    env = subprocess_env()
    try:
        p = subprocess.run(
            ("cset", "recipe-id", "--recipe", file),
            capture_output=True,
            check=True,
            env=env,
        )
    except subprocess.CalledProcessError as err:
        logging.exception(
            "cset recipe-id exited with non-zero code %s.\nstdout: %s\nstderr: %s",
            err.returncode,
            # Presume that subprocesses have the same IO encoding as this one.
            # Honestly, on all our supported platforms this will be "utf-8".
            err.stdout.decode(sys.stdout.encoding),
            err.stderr.decode(sys.stderr.encoding),
        )
        raise
    recipe_id = p.stdout.decode(sys.stdout.encoding).strip()
    model_identifiers = sorted(os.environ["MODEL_IDENTIFIERS"].split())
    return f"m{'_m'.join(model_identifiers)}_{recipe_id}"


def output_directory():
    """Get the plot output directory for the recipe."""
    share_directory = os.environ["CYLC_WORKFLOW_SHARE_DIR"]
    cycle_point = os.environ["CYLC_TASK_CYCLE_POINT"]
    return f"{share_directory}/web/plots/{recipe_id()}_{cycle_point}"


def data_directories() -> list[str]:
    """Get the input data directories for the cycle."""
    model_identifiers = sorted(os.environ["MODEL_IDENTIFIERS"].split())
    if os.getenv("DO_CASE_AGGREGATION"):
        cylc_workflow_share_dir = os.environ["CYLC_WORKFLOW_SHARE_DIR"]
        return [
            f"{cylc_workflow_share_dir}/cycle/*/data/{model_id}"
            for model_id in model_identifiers
        ]
    else:
        rose_datac = os.environ["ROSE_DATAC"]
        return [f"{rose_datac}/data/{model_id}" for model_id in model_identifiers]


def create_diagnostic_archive(output_directory):
    """Create archive for easy download of plots and data."""
    output_directory = Path(output_directory)
    archive_path = output_directory / "diagnostic.zip"
    with zipfile.ZipFile(
        archive_path, "w", compression=zipfile.ZIP_DEFLATED
    ) as archive:
        for file in output_directory.rglob("*"):
            # Check the archive doesn't add itself.
            if not file.samefile(archive_path):
                archive.write(file, arcname=file.relative_to(output_directory))


def run_recipe_steps():
    """Process data and produce output plots."""
    command = (
        ["cset", "bake", "--recipe", recipe_file(), "--input-dir"]
        + data_directories()
        + ["--output-dir", output_directory()]
    )

    colorbar_file = os.getenv("COLORBAR_FILE")
    if colorbar_file:
        command.append(f"--style-file={colorbar_file}")

    plot_resolution = os.getenv("PLOT_RESOLUTION")
    if plot_resolution:
        command.append(f"--plot-resolution={plot_resolution}")

    logging.info("Running %s", " ".join(command))
    try:
        subprocess.run(command, check=True, env=subprocess_env(), capture_output=True)
    except subprocess.CalledProcessError as err:
        logging.exception(
            "cset bake exited with non-zero code %s.\nstdout: %s\nstderr: %s",
            err.returncode,
            err.stdout.decode(sys.stdout.encoding),
            err.stderr.decode(sys.stderr.encoding),
        )
        raise
    create_diagnostic_archive(output_directory())


def run():
    """Run workflow script."""
    run_recipe_steps()



# © Crown copyright, Met Office (2022-2025) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Common operator functionality used across CSET.

Functions below should only be added if it is not suitable as a standalone
operator, and will be used across multiple operators.
"""

import logging

import iris
import iris.cube
import iris.exceptions
import iris.util


def get_cube_yxcoordname(cube: iris.cube.Cube) -> tuple[str, str]:
    """
    Return horizontal coordinate name(s) from a given cube.

    Arguments
    ---------

    cube: iris.cube.Cube
        An iris cube which will be checked to see if it contains coordinate
        names that match a pre-defined list of acceptable horizontal
        coordinate names.

    Returns
    -------
    (y_coord, x_coord)
        A tuple containing the horizontal coordinate name for latitude and longitude respectively
        found within the cube.

    Raises
    ------
    ValueError
        If a unique y/x horizontal coordinate cannot be found.
    """
    # Acceptable horizontal coordinate names.
    X_COORD_NAMES = ["longitude", "grid_longitude", "projection_x_coordinate", "x"]
    Y_COORD_NAMES = ["latitude", "grid_latitude", "projection_y_coordinate", "y"]

    # Get a list of coordinate names for the cube
    coord_names = [coord.name() for coord in cube.coords()]

    # Check which x-coordinate we have, if any
    x_coords = [coord for coord in coord_names if coord in X_COORD_NAMES]
    if len(x_coords) != 1:
        raise ValueError("Could not identify a unique x-coordinate in cube")

    # Check which y-coordinate we have, if any
    y_coords = [coord for coord in coord_names if coord in Y_COORD_NAMES]
    if len(y_coords) != 1:
        raise ValueError("Could not identify a unique y-coordinate in cube")

    return (y_coords[0], x_coords[0])


def is_spatialdim(cube: iris.cube.Cube) -> bool:
    """Determine whether a cube is has two spatial dimension coordinates.

    If cube has both spatial dims, it will contain two unique coordinates
    that explain space (latitude and longitude). The coordinates have to
    be iterable/contain usable dimension data, as cubes may contain these
    coordinates as scalar dimensions after being collapsed.

    Arguments
    ---------
    cube: iris.cube.Cube
        An iris cube which will be checked to see if it contains coordinate
        names that match a pre-defined list of acceptable coordinate names.

    Returns
    -------
    bool
        If true, then the cube has a spatial projection and thus can be plotted
        as a map.
    """
    # Acceptable horizontal coordinate names.
    X_COORD_NAMES = ["longitude", "grid_longitude", "projection_x_coordinate", "x"]
    Y_COORD_NAMES = ["latitude", "grid_latitude", "projection_y_coordinate", "y"]

    # Get a list of coordinate names for the cube
    coord_names = [coord.name() for coord in cube.dim_coords]
    x_coords = [coord for coord in coord_names if coord in X_COORD_NAMES]
    y_coords = [coord for coord in coord_names if coord in Y_COORD_NAMES]

    # If there is one coordinate for both x and y direction return True.
    if len(x_coords) == 1 and len(y_coords) == 1:
        return True
    else:
        return False


def is_transect(cube: iris.cube.Cube) -> bool:
    """Determine whether a cube is a transect.

    If cube is a transect, it will contain only one spatial (map) coordinate,
    and one vertical coordinate (either pressure or model level).

    Arguments
    ---------
    cube: iris.cube.Cube
        An iris cube which will be checked to see if it contains coordinate
        names that match a pre-defined list of acceptable coordinate names.

    Returns
    -------
    bool
        If true, then the cube is a transect that contains one spatial (map)
        coordinate and one vertical coordinate.
    """
    # Acceptable spatial (map) coordinate names.
    SPATIAL_MAP_COORD_NAMES = [
        "longitude",
        "grid_longitude",
        "projection_x_coordinate",
        "x",
        "latitude",
        "grid_latitude",
        "projection_y_coordinate",
        "y",
        "distance",
    ]

    # Acceptable vertical coordinate names
    VERTICAL_COORD_NAMES = ["pressure", "model_level_number", "level_height"]

    # Get a list of coordinate names for the cube
    coord_names = [coord.name() for coord in cube.coords(dim_coords=True)]

    # Check which spatial coordinates we have.
    spatial_coords = [
        coord for coord in coord_names if coord in SPATIAL_MAP_COORD_NAMES
    ]
    if len(spatial_coords) != 1:
        return False

    # Check which vertical coordinates we have.
    vertical_coords = [coord for coord in coord_names if coord in VERTICAL_COORD_NAMES]
    if len(vertical_coords) != 1:
        return False

    # Passed criteria so return True
    return True


def fully_equalise_attributes(cubes: iris.cube.CubeList):
    """Remove any unique attributes between cubes or coordinates in place."""
    # Equalise cube attributes.
    removed = iris.util.equalise_attributes(cubes)
    logging.debug("Removed attributes from cube: %s", removed)

    # Equalise coordinate attributes.
    coord_sets = [{coord.name() for coord in cube.coords()} for cube in cubes]

    all_coords = set.union(*coord_sets)
    coords_to_equalise = set.intersection(*coord_sets)
    coords_to_remove = set.difference(all_coords, coords_to_equalise)

    logging.debug("All coordinates: %s", all_coords)
    logging.debug("Coordinates to remove: %s", coords_to_remove)
    logging.debug("Coordinates to equalise: %s", coords_to_equalise)

    for coord in coords_to_remove:
        for cube in cubes:
            try:
                cube.remove_coord(coord)
                logging.debug("Removed coordinate %s from %s cube.", coord, cube.name())
            except iris.exceptions.CoordinateNotFoundError:
                pass

    for coord in coords_to_equalise:
        removed = iris.util.equalise_attributes([cube.coord(coord) for cube in cubes])
        logging.debug("Removed attributes from coordinate %s: %s", coord, removed)

    return cubes


def is_time_aggregatable(cube: iris.cube.Cube) -> bool:
    """Determine whether a cube can be aggregated in time.

    If a cube is aggregatable it will contain both a 'forecast_reference_time'
    and 'forecast_period' coordinate as dimensional coordinates.

    Arguments
    ---------
    cube: iris.cube.Cube
        An iris cube which will be checked to see if it is aggregatable based
        on a set of pre-defined dimensional time coordinates:
        'forecast_period' and 'forecast_reference_time'.

    Returns
    -------
    bool
        If true, then the cube is aggregatable and contains dimensional
        coordinates including both 'forecast_reference_time' and
        'forecast_period'.
    """
    # Acceptable time coordinate names for aggregatable cube.
    TEMPORAL_COORD_NAMES = ["forecast_period", "forecast_reference_time"]

    # Coordinate names for the cube.
    coord_names = [coord.name() for coord in cube.coords(dim_coords=True)]

    # Check which temporal coordinates we have.
    temporal_coords = [coord for coord in coord_names if coord in TEMPORAL_COORD_NAMES]
    # Return whether both coordinates are in the temporal coordinates.
    return len(temporal_coords) == 2



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Operators to perform various kind of filtering."""

import logging

import iris
import iris.cube
import iris.exceptions
import numpy as np


def apply_mask(
    original_field: iris.cube.Cube,
    mask: iris.cube.Cube,
) -> iris.cube.Cube:
    """Apply a mask to given data as a masked array.

    Parameters
    ----------
    original_field: iris.cube.Cube
        The field to be masked.
    mask: iris.cube.Cube
        The mask being applied to the original field.

    Returns
    -------
    masked_field: iris.cube.Cube
        A cube of the masked field.

    Notes
    -----
    The mask is first converted to 1s and NaNs before multiplication with
    the original data.

    As discussed in generate_mask, you can combine multiple masks in a
    recipe using other functions before applying the mask to the data.

    Examples
    --------
    >>> land_points_only = apply_mask(temperature, land_mask)
    """
    # Ensure mask is only 1s or NaNs.
    mask.data[mask.data == 0] = np.nan
    mask.data[~np.isnan(mask.data)] = 1
    logging.info(
        "Mask set to 1 or 0s, if addition of multiple masks results"
        "in values > 1 these are set to 1."
    )
    masked_field = original_field.copy()
    masked_field.data *= mask.data
    masked_field.attributes["mask"] = f"mask_of_{original_field.name()}"
    return masked_field


def filter_cubes(
    cube: iris.cube.Cube | iris.cube.CubeList,
    constraint: iris.Constraint,
    **kwargs,
) -> iris.cube.Cube:
    """Filter a CubeList down to a single Cube based on a constraint.

    Arguments
    ---------
    cube: iris.cube.Cube | iris.cube.CubeList
        Cube(s) to filter
    constraint: iris.Constraint
        Constraint to extract

    Returns
    -------
    iris.cube.Cube

    Raises
    ------
    ValueError
        If the constraint doesn't produce a single cube.
    """
    filtered_cubes = cube.extract(constraint)
    # Return directly if already a cube.
    if isinstance(filtered_cubes, iris.cube.Cube):
        return filtered_cubes
    # Check filtered cubes is a CubeList containing one cube.
    if isinstance(filtered_cubes, iris.cube.CubeList) and len(filtered_cubes) == 1:
        return filtered_cubes[0]
    else:
        raise ValueError(
            f"Constraint doesn't produce single cube. Constraint: {constraint}"
            f"\nSource: {cube}\nResult: {filtered_cubes}"
        )


def filter_multiple_cubes(
    cubes: iris.cube.Cube | iris.cube.CubeList,
    **kwargs,
) -> iris.cube.CubeList:
    """Filter a CubeList on multiple constraints, returning another CubeList.

    Arguments
    ---------
    cube: iris.cube.Cube | iris.cube.CubeList
        Cube(s) to filter
    constraint: iris.Constraint
        Constraint to extract. This must be a named argument. There can be any
        number of additional constraints, they just need unique names.

    Returns
    -------
    iris.cube.CubeList

    Raises
    ------
    ValueError
        The constraints don't produce a single cube per constraint.
    """
    # Ensure input is a CubeList.
    if isinstance(cubes, iris.cube.Cube):
        cubes = iris.cube.CubeList((cubes,))
    if len(kwargs) < 1:
        raise ValueError("Must have at least one constraint.")
    try:
        filtered_cubes = cubes.extract_cubes(kwargs.values())
    except iris.exceptions.ConstraintMismatchError as err:
        raise ValueError(
            "The constraints don't produce a single cube per constraint."
        ) from err
    return filtered_cubes


def generate_mask(
    mask_field: iris.cube.Cube,
    condition: str,
    value: float,
) -> iris.cube.Cube:
    """Generate a mask to remove data not meeting conditions.

    Parameters
    ----------
    mask_field: iris.cube.Cube
        The field to be used for creating the mask.
    condition: str
        The type of condition applied, six available options:
        '==','!=','<','<=','>', and '>='.
    value: float
        The value on the right hand side of the condition.

    Returns
    -------
    mask: iris.cube.Cube
        Mask meeting the condition applied.

    Raises
    ------
    ValueError: Unexpected value for condition. Expected ==, !=, >, >=, <, <=.
                Got {condition}.
        Raised when condition is not supported.

    Notes
    -----
    The mask is created in the opposite sense to numpy.ma.masked_arrays. This
    method was chosen to allow easy combination of masks together outside of
    this function using misc.addition or misc.multiplication depending on
    applicability. The combinations can be of any fields such as orography >
    500 m, and humidity == 100 %.

    The conversion to a masked array occurs in the apply_mask routine, which
    should happen after all relevant masks have been combined.

    Examples
    --------
    >>> land_mask = generate_mask(land_sea_mask,'==',1)
    """
    mask = mask_field.copy()
    mask.data = np.zeros(mask.data.shape)
    match condition:
        case "==":
            mask.data[mask_field.data == value] = 1
        case "!=":
            mask.data[mask_field.data != value] = 1
        case ">":
            mask.data[mask_field.data > value] = 1
        case ">=":
            mask.data[mask_field.data >= value] = 1
        case "<":
            mask.data[mask_field.data < value] = 1
        case "<=":
            mask.data[mask_field.data <= value] = 1
        case _:
            raise ValueError("""Unexpected value for condition. Expected ==, !=,
                              >, >=, <, <=. Got {condition}.""")
    mask.attributes["mask"] = f"mask_for_{mask_field.name()}_{condition}_{value}"
    return mask
